
from tests import helpers
from datetime import datetime, timezone
from joule.services import parse_pipe_config
from joule.models import DataStream, folder, Folder, EventStream
from joule.errors import ConfigurationError


class TestFolder(helpers.DbTestCase):

    def test_get_stream_path(self):
        # create a stream and then find it again
        stream = parse_pipe_config.run("/very/long/path/to/stream:float32[x]", self.db)
        path = folder.get_stream_path(stream)
        self.assertEqual("/very/long/path/to/stream", path)

        # stream outside the database has no path
        stream = DataStream(name="all alone")
        self.assertIsNone(folder.get_stream_path(stream))

        # remote streams are [node] /path/to/stream
        stream = DataStream(name="remote_stream")
        stream.set_remote("remote_node", "/path/to/stream")
        path = folder.get_stream_path(stream)
        self.assertEqual("[remote_node] /path/to/stream", path)

    def test_find_or_create(self):
        folder.find("/new/folder/path", self.db, create=True)
        self.assertEqual(self.db.query(Folder).count(), 4)
        # trailing slash raises an error
        with self.assertRaisesRegex(ConfigurationError, 'invalid path'):
            folder.find("/new/folder/path/", self.db, create=True)
        # does not create folder if create flag is false
        result = folder.find("/new/different/path", self.db)
        self.assertIsNone(result)
        self.assertEqual(self.db.query(Folder).count(), 4)
        # raises error if the requested stream_type is not valid
        with self.assertRaisesRegex(ValueError, 'type'):
            folder.find_stream_by_path("/new/folder/path", self.db, stream_type="invalid")

    def test_updates_attributes(self):
        my_folder = folder.find("/new/folder/path", self.db, create=True)
        my_folder.update_attributes({"name": "new name", "description": "new description"})
        self.assertEqual(my_folder.name, "new name")
        self.assertEqual(my_folder.description, "new description")
        # validates name attribute
        for name in ["invalid/name", "", None]:
            with self.assertRaises(ConfigurationError):
                my_folder.update_attributes({"name": name})


    def test_contains_streams(self):
        my_folder = folder.find("/new/folder/path", self.db, create=True)
        my_folder.data_streams = [helpers.create_stream("stream1", "int16_2")]
        self.db.add(my_folder)
        self.db.commit()
        f = folder.find("/new", self.db)
        self.assertTrue(f.contains_streams())
        self.assertTrue(my_folder.contains_streams())

        my_folder = folder.find("/events/new/folder/path", self.db, create=True)
        my_folder.event_streams = [EventStream(name="event_stream1", updated_at=datetime.now(timezone.utc))]
        self.db.add(my_folder)
        self.db.commit()
        f = folder.find("/events", self.db)
        self.assertTrue(f.contains_streams())
        self.assertTrue(my_folder.contains_streams())

        empty_folder = folder.find("/empty/new/folder", self.db, create=True)
        self.db.add(empty_folder)
        self.db.commit()
        f = folder.find("/empty", self.db)
        self.assertFalse(f.contains_streams())
        self.assertFalse(empty_folder.contains_streams())

    def test_locked(self):
        my_folder = folder.find("/new/folder/path", self.db, create=True)
        my_stream = helpers.create_stream("stream1", "int16_2")
        my_stream.is_configured = True
        my_stream.folder = my_folder
        self.db.add(my_stream)
        self.db.commit()
        f = folder.find("/new", self.db)
        self.assertTrue(f.locked)
        self.assertTrue(my_folder.locked)

        my_stream.is_configured = False
        self.db.commit()
        self.assertFalse(f.locked)
        self.assertFalse(my_folder.locked)

        empty_folder = folder.find("/empty/new/folder", self.db, create=True)
        self.db.add(empty_folder)
        self.db.commit()
        f = folder.find("/empty", self.db)
        self.assertFalse(f.locked)
        self.assertFalse(empty_folder.locked)







from sqlalchemy import create_engine
from sqlalchemy.orm import Session
from typing import List, Dict, Tuple
import numpy as np
import asyncio
import argparse
import psycopg2
import testing.postgresql

from unittest import mock
from typing import Optional, Callable, Coroutine

from joule.models.data_store.errors import DataError, InsufficientDecimationError
from joule.models.data_store.event_store import EventStore
from joule.models import folder
from joule.models.data_store.event_store import StreamInfo as EventStreamInfo
from joule.api.event_stream import EventStream as ApiEventStream
from joule.models import (Base, DataStore, DataStream,
                          StreamInfo, DbInfo, pipes,
                          worker, EventStream)
from joule.models.event_stream import from_json as event_stream_from_json
from joule.errors import SubscriptionError, EmptyPipeError
from joule.services import parse_pipe_config
from tests import helpers

Loop = asyncio.AbstractEventLoop


def create_db(pipe_configs: List[str], 
              api_event_streams: Optional[List[ApiEventStream]]=None,
              event_path="/events") -> Tuple[Session, testing.postgresql.Postgresql]:
    postgresql = testing.postgresql.Postgresql()
    db_url = postgresql.url()
    conn = psycopg2.connect(db_url)
    cur = conn.cursor()
    cur.execute("CREATE SCHEMA metadata")
    conn.commit()
    conn.close()
    engine = create_engine(db_url)
    Base.metadata.create_all(engine)
    db = Session(bind=engine)
    for pipe_config in pipe_configs:
        parse_pipe_config.run(pipe_config, db)
    db.commit()
    if api_event_streams is None:
        return db, postgresql
    event_folder =  folder.find(event_path, db, create=True)

    for api_event_stream in api_event_streams or []:
        event_stream = event_stream_from_json(api_event_stream.to_json())
        event_folder.event_streams.append(event_stream)
        event_stream.touch()
    db.commit()
    return db, postgresql


class MockEventStore(EventStore):
    def __init__(self):
        self.reset()

    def reset(self):
        self.stream_info = {}
        self.destroyed_stream_id = None
        self.create_called = False
        self.new_stream = None
        self.upsert_called = False
        self.upserted_events = []
        self.count_called = False
        self.count_params = None
        self.extract_params = None
        self.extract_called = False
        self.histogram_called = False
        self.remove_called = False
        self._events = []

    async def initialize(self, pool=None):
        # API compatibility with DataStore, does not require any initialization
        pass

    def set_events(self, events):
        self._events = events

    async def info(self, streams: List[EventStream]) -> Dict[int, EventStreamInfo]:
        info_dict = {}
        for s in streams:
            if s in self.stream_info:
                info_dict[s.id] = self.stream_info[s]
        return info_dict

    async def destroy(self, stream: EventStream):
        self.destroyed_stream_id = stream.id

    async def create(self, new_stream):
        self.new_stream = new_stream
        self.create_called = True

    async def upsert(self, stream, events):
        self.upserted_events = events
        self.upsert_called = True
        return events
    
    async def count(self, stream, start=None, end=None, json_filter=None,include_on_going_events=False):
        self.count_called = True
        self.count_params = (stream, start, end, json_filter, include_on_going_events)
        return len(self._events)
    
    async def extract(self, stream, start=None, end=None, json_filter=None, limit=0, include_on_going_events=False):
        self.extract_params = (stream, start, end, json_filter, limit, include_on_going_events)
        self.extract_called = True
        return self._events[:limit]
    
    async def histogram(self, stream, start=None, end=None, json_filter=None, nbuckets=0):
        self.histogram_called = True
        self.histogram_params = (stream, start, end, json_filter, nbuckets)
        return [] # empty histogram
    
    async def remove(self, stream, start=None, end=None, json_filter=None):
        self.remove_called = True
        self.remove_params = (stream, start, end, json_filter)

class MockStore(DataStore):
    def __init__(self):
        self.stream_info = {}
        self.supports_decimation_management = True # act like TimescaleDB
        # stubs to track data controller execution
        self.nchunks = 3
        self.nintervals = 4
        self.raise_data_error = False
        self.raise_decimation_error = False
        self.no_data = False
        self.raise_data_error = False
        self.inserted_data = False
        self.removed_data_bounds = (None, None)
        self.destroyed_stream_id = None
        self.raise_data_error = False
        self.dropped_decimations = False
        self.redecimated_data = False
        self.merge_gap = None
        self.consolidate_called = False
        self.consolidate_params = None

    async def initialize(self, streams: List[DataStream]):
        # API compatibility with DataStore, does not require any initialization
        pass

    async def insert(self, stream: DataStream,
                     data: np.ndarray, start: int, end: int):
        # API compatibility with DataStore, no action required
        pass

    async def spawn_inserter(self, stream: DataStream, pipe: pipes.InputPipe, insert_period=None, merge_gap=0) -> asyncio.Task:
        async def task():
            if self.raise_data_error:
                raise DataError("test error")
            self.inserted_data = True
            self.merge_gap = merge_gap

        return asyncio.create_task(task())

    def configure_extract(self, nchunks, nintervals=4,
                          decimation_error=False,
                          data_error=False,
                          no_data=False):
        self.nchunks = nchunks
        assert nintervals >= 4
        self.nintervals = nintervals
        self.raise_decimation_error = decimation_error
        self.raise_data_error = data_error
        self.no_data = no_data

    async def extract(self, stream: DataStream, start: Optional[int], end: Optional[int],
                      callback: Callable[[np.ndarray, str, bool], Coroutine],
                      max_rows: int = None, decimation_level=1):
        if self.no_data:
            return  # do not call the callback func
        if self.raise_data_error:
            raise DataError("backend error")
        if self.raise_decimation_error:
            raise InsufficientDecimationError("insufficient decimation")
        if decimation_level is not None and decimation_level > 1:
            layout = stream.decimated_layout
        else:
            decimation_level = 1
            layout = stream.layout
        # We always send at least 4 intervals to cover all the cases
        assert self.nintervals >= 4
        first_half = self.nintervals // 2
        second_half = self.nintervals - first_half
        ### This tries to simulate the behavior of the extract function in the data store
        ### with multiple network conditions and store versions
        ###
        # CASE 1: send the data in multiple batches with interval tokens separate from the data (old behavior)
        for i in range(first_half-1): 
            for _ in range(self.nchunks):
                await callback(helpers.create_data(layout, length=25), layout, decimation_level)
            if i < (self.nintervals - 1):
                await callback(pipes.interval_token(layout), layout, decimation_level)
        # CASE 2: send the data in multiple batches with interval tokens at the end of the data (new behavior)
        for i in range(second_half-1):
            for _ in range(self.nchunks-1):
                await callback(helpers.create_data(layout, length=25), layout, decimation_level)
            last_chunk = helpers.create_data(layout, length=25)
            if i < (self.nintervals - 1): # put the token at the end of the chunk
                data_with_token = np.hstack((last_chunk,pipes.interval_token(layout)))
                await callback(data_with_token, layout, decimation_level)
            else:
                await callback(last_chunk, layout, decimation_level)
        # CASE 3: send an entire interval at once with the token at the end
        full_data_with_token = np.hstack((helpers.create_data(layout, length=25*self.nchunks),
                                          pipes.interval_token(layout)))
        await callback(full_data_with_token, layout, decimation_level)
        # Last interval has no token
        full_data_no_token = helpers.create_data(layout, length=25*self.nchunks)
        await callback(full_data_no_token, layout, decimation_level)


    async def intervals(self, stream: 'DataStream', start: Optional[int], end: Optional[int]):
        if start is None:
            if end is None:
                # just a random set of intervals
                return [[0, 100], [200, 300]]
            else:
                return [[end - 10, end]]
        else:
            if end is None:
                return [[start, start + 10]]
            else:
                return [[start, end]]

    async def drop_decimations(self, stream: 'DataStream'):
        self.dropped_decimations = True

    async def decimate(self, stream: 'DataStream'):
        self.redecimated_data = True

    async def remove(self, stream: DataStream, start: Optional[int], end: Optional[int]):
        self.removed_data_bounds = (start, end)

    async def destroy(self, stream: DataStream):
        self.destroyed_stream_id = stream.id

    async def consolidate(self, stream: 'DataStream', start: Optional[int], end: Optional[int], max_gap: int) -> int:
        return 0

    async def destroy_all(self):
        raise Exception("not implemented!")

    async def info(self, streams: List[DataStream]) -> Dict[int, StreamInfo]:
        info_dict = {}
        for s in streams:
            if s in self.stream_info:
                info_dict[s.id] = self.stream_info[s]
        return info_dict

    async def dbinfo(self) -> DbInfo:
        return DbInfo('/file/path', 0, 0, 0, 0)

    async def consolidate(self, stream: 'DataStream', start: Optional[int], end: Optional[int], max_gap: int) -> int:
        self.consolidate_called = True
        self.consolidate_params = (stream, start, end, max_gap)
        return 0

    def close(self):
        pass

    # -- special mock tools --
    def set_info(self, stream: DataStream, info: StreamInfo):
        self.stream_info[stream] = info


class MockWorker:
    def __init__(self, name, inputs, outputs, uuid=1, is_app=False, socket=None):
        self.name = name
        self.description = "description for %s" % name
        self.is_app = is_app
        self.uuid = uuid
        self.app_socket = socket
        self.input_connections = []
        self.workers = []
        self.proxies = []
        for (name, path) in inputs.items():
            self.input_connections.append(argparse.Namespace(name=name, location=path))
        self.output_connections = []
        for (name, path) in outputs.items():
            self.output_connections.append(argparse.Namespace(name=name, location=path))

    async def statistics(self) -> worker.Statistics:
        return worker.Statistics(100, 100, 1.0, 100)

    @property
    def logs(self):
        return ["log entry1", "log entry2"]


class MockSupervisor:
    def __init__(self):
        self.subscription_pipe = None
        self.subscribed_stream = None
        self.raise_error = False
        # if True do not close the pipe (simulate an ongoing transaction)
        self.hang_pipe = False
        self.unsubscribe_mock = mock.Mock()
        self.unsubscribe_calls = 0

    def get_module_socket(self, uuid):
        for worker in self.workers:
            if worker.uuid == uuid:
                return f"--{uuid}--"
        return None
    
    def get_proxy_url(self, uuid):
        for proxy in self.proxies:
            if proxy.uuid == uuid:
                return proxy.url
        return None
    
    def unsubscribe(self):
        self.unsubscribe_calls += 1

    def subscribe(self, stream: DataStream, pipe: pipes.LocalPipe):
        if self.raise_error:
            raise SubscriptionError()

        self.subscribed_stream = stream
        while True:
            try:
                data = self.subscription_pipe.read_nowait()
                self.subscription_pipe.consume(len(data))
                pipe.write_nowait(data)
            except EmptyPipeError:
                if not self.hang_pipe:
                    pipe.close_nowait()
                break
            if self.subscription_pipe.end_of_interval:
                pipe.close_interval_nowait()
        # return a mock as the unsubscribe callback
        return self.unsubscribe



import unittest
from joule import api, errors


class TestFolderMethods(unittest.IsolatedAsyncioTestCase):

    async def asyncSetUp(self):
        self.node = api.get_node()
        """
        └── test
            ├── f1
            │   ├── f1a
            │   │   ├── s1a1
            │   │   └── s1a2
            │   └── s1a
            └── f2
                └── f2a
                    └── s2a1a
        """
        stream = api.DataStream(name="s1a1", elements=[api.Element(name="e1")])
        await self.node.data_stream_create(stream, "/test/f1/f1a")
        stream.name = "s1a2"
        await self.node.data_stream_create(stream, "/test/f1/f1a")
        stream.name = "s1a"
        await self.node.data_stream_create(stream, "/test/f1")
        stream.name = "s2a1a"
        await self.node.data_stream_create(stream, "/test/f2/f2a")

    async def asyncTearDown(self):
        await self.node.folder_delete("/test", recursive=True)
        await self.node.close()

    async def test_folder_root(self):
        root = await self.node.folder_root()
        self.assertTrue(root.locked)
        self.assertEqual('root', root.name)
        # should have all of the streams and folders, check a few
        f_test = [f for f in root.children if f.name == 'test'][0]
        f1 = [f for f in f_test.children if f.name == 'f1'][0]
        s1a = [s for s in f1.data_streams if s.name == 's1a'][0]
        self.assertEqual("e1", s1a.elements[0].name)

    async def test_folder_move_to_existing_destination(self):
        await self.node.folder_move("/test/f1/f1a", "/test/f2")
        s1a1 = await self.node.data_stream_get("/test/f2/f1a/s1a1")
        self.assertEqual("s1a1", s1a1.name)

    async def test_folder_move_to_new_destination(self):
        await self.node.folder_move("/test/f1/f1a", "/test/new")
        s1a1 = await self.node.data_stream_get("/test/new/f1a/s1a1")
        self.assertEqual("s1a1", s1a1.name)

    async def test_folder_move_errors(self):
        # source does not exist
        with self.assertRaises(errors.ApiError):
            await self.node.folder_move("/test/f1/bad", "/test/f2")
        # circular graph
        with self.assertRaises(errors.ApiError):
            await self.node.folder_move("/test/f1", "/test/f1/f1a")
        # name conflict
        stream = api.DataStream(name="s1a1", elements=[api.Element(name="e1")])
        await self.node.data_stream_create(stream, "/test/f1/f2")
        with self.assertRaises(errors.ApiError):
            await self.node.folder_move("/test/f1/f2", "/test")
        # locked folder
        with self.assertRaises(errors.ApiError):
            await self.node.folder_move("/live/base", "/test")

    async def test_folder_update(self):
        f1 = await self.node.folder_get("/test/f1")
        f1.name = "f1_new"
        f1.description = "new description"
        await self.node.folder_update(f1)
        f1_new = await self.node.folder_get(f1.id)
        self.assertEqual(f1_new.name, "f1_new")
        self.assertEqual(f1_new.description, "new description")

    async def test_folder_update_errors(self):
        # name cannot be blank
        f1 = await self.node.folder_get("/test/f1")
        f1.name = ""
        with self.assertRaises(errors.ApiError):
            await self.node.folder_update(f1)
        f1.name = None
        with self.assertRaises(errors.ApiError):
            await self.node.folder_update(f1)
        # name cannot conflict with a peer
        f1.name = "f2"
        with self.assertRaises(errors.ApiError):
            await self.node.folder_update(f1)
        # make sure name didn't change
        f1 = await self.node.folder_get(f1.id)
        self.assertEqual("f1", f1.name)

    async def test_folder_delete(self):
        s2a1a = await self.node.data_stream_get("/test/f2/f2a/s2a1a")
        await self.node.folder_delete("/test/f2/f2a")
        # folder is deleted
        f2 = await self.node.folder_get("/test/f2")
        self.assertEqual(0, len(f2.children))
        # streams in folder are deleted
        with self.assertRaises(errors.ApiError):
            await self.node.data_stream_get(s2a1a)

    async def test_folder_delete_errors(self):
        # folder does not exist
        with self.assertRaises(errors.ApiError):
            await self.node.folder_delete("/test/bad")
        # non-empty folder without recursive flag
        with self.assertRaises(errors.ApiError):
            await self.node.folder_delete("/test/f2")
        # locked folder
        with self.assertRaises(errors.ApiError):
            await self.node.folder_delete("/live")
        with self.assertRaises(errors.ApiError):
            await self.node.folder_delete("/archive")
        # all streams should still exist
        await self.node.data_stream_get("/live/base")
        await self.node.data_stream_get("/archive/data1")
        await self.node.data_stream_get("/test/f2/f2a/s2a1a")




import asyncio


class QueueReader:
    def __init__(self, delay=0.0):
        self._queue = asyncio.Queue()
        self.delay = delay

    async def put(self, data):
        await self._queue.put(data)

    async def read(self, _: int):
        if self.delay > 0:
            await asyncio.sleep(self.delay)
        return await self._queue.get()

    def at_eof(self):
        return self._queue.empty()




#!/usr/bin/env python3
import time
import io
import unittest
import basic


def main():
    loader = unittest.TestLoader()
    suite = unittest.TestSuite()
    suite.addTests(loader.loadTestsFromModule(basic))
    output = io.StringIO()
    runner = unittest.TextTestRunner(stream=output, failfast=True)
    result = runner.run(suite)
    if result.wasSuccessful():
        print("OK")
    else:
        print("FAIL")
        output.seek(0)
        print(output.read())


if __name__ == '__main__':
    time.sleep(1)
    main()



from tests.helpers import DbTestCase
from joule import update_users
from joule.models.master import Master
import tempfile

class TestUpdateUsers(DbTestCase):

   

    def test_adds_users_from_file(self):
        with tempfile.NamedTemporaryFile() as user_file:
            # 1. Create two users
            with open(user_file.name, 'w') as f:
                f.write("""
                # add two users
                johndoe, abcdefghijkl123
                # comments and blank lines should be ignored

                janedoe, 123abcdefghijkl
                """)
            update_users.update_users_from_file(user_file.name, self.db)
            # make sure both users are in the database
            self.assertEqual(self.db.query(Master).count(), 2)
            user1: Master = self.db.query(Master).filter_by(name="johndoe").one()
            user2: Master = self.db.query(Master).filter_by(name="janedoe").one()
            self.assertEqual(user1.key, "abcdefghijkl123")
            self.assertEqual(user1.type, Master.TYPE.USER)
            self.assertEqual(user2.key, "123abcdefghijkl")
            self.assertEqual(user2.type, Master.TYPE.USER)

            # 2. Update second user's API key
            with open(user_file.name, 'w') as f:
                f.write("""
                # change second user's key
                johndoe, abcdefghijkl123
                janedoe, 123zxyabcdefghijkl
                """)
            # should get a log message that johndoe is unchanged
            with self.assertLogs(level='INFO') as logs:
                update_users.update_users_from_file(user_file.name, self.db)
            log_output = " ".join(logs.output)
            self.assertIn("johndoe already exists with this key", log_output)
            self.assertIn("Modifying API key for janedoe", log_output)
            update_users.update_users_from_file(user_file.name, self.db)
            user1: Master = self.db.query(Master).filter_by(name="johndoe").one()
            user2: Master = self.db.query(Master).filter_by(name="janedoe").one()
            self.assertEqual(user1.key, "abcdefghijkl123")
            self.assertEqual(user1.type, Master.TYPE.USER)
            self.assertEqual(user2.key, "123zxyabcdefghijkl")
            self.assertEqual(user2.type, Master.TYPE.USER)
            
            # 3. Add two new users, remove first user
            with open(user_file.name, 'w') as f:
                f.write("""
                # add two new users, remove first
                DELETE johndoe
                newuser1, 456zxyabcdefghijkl
                newuser2, 789zxyabcdefghijkl
                """)
            update_users.update_users_from_file(user_file.name, self.db)
            self.assertEqual(self.db.query(Master).count(), 3)
            # make sure johndoe is deleted
            with self.assertRaises(Exception):
                self.db.query(Master).filter_by(name="johndoe").one()
            user3: Master = self.db.query(Master).filter_by(name="newuser1").one()
            user4: Master = self.db.query(Master).filter_by(name="newuser2").one()
            self.assertEqual(user3.key, "456zxyabcdefghijkl")
            self.assertEqual(user4.key, "789zxyabcdefghijkl")

            # 4. Remove users with a LIKE clause
            with open(user_file.name, 'w') as f:
                f.write("""
                # remove users with LIKE clause
                DELETE LIKE newuser%
                """)
            update_users.update_users_from_file(user_file.name, self.db)
            self.assertEqual(self.db.query(Master).count(), 1)
            with self.assertRaises(Exception):
                self.db.query(Master).filter_by(name="newuser1").one()
            with self.assertRaises(Exception):
                self.db.query(Master).filter_by(name="newuser2").one()
            # there should only be one user left
            user2: Master = self.db.query(Master).filter_by(name="janedoe").one()
            self.assertEqual(user2.key, "123zxyabcdefghijkl")
            self.assertEqual(user2.type, Master.TYPE.USER)

    def test_errors_on_adds_users_from_file(self):
        with tempfile.NamedTemporaryFile() as user_file:
            # create two users
            with open(user_file.name, 'w') as f:
                f.write("""
                # add two users
                johndoe, abcdefghijkl123
                janedoe, 123abcdefghijkl
                """)
            update_users.update_users_from_file(user_file.name, self.db)
            # make sure both users are in the database
            self.assertEqual(self.db.query(Master).count(), 2)

        # 1. invalid syntax
        with tempfile.NamedTemporaryFile() as user_file:
            with open(user_file.name, 'w') as f:
                f.write("""
                this is an error
                """)
            with self.assertLogs(level='ERROR'):
                update_users.update_users_from_file(user_file.name, self.db)
            # make sure both users are in the database
            self.assertEqual(self.db.query(Master).count(), 2)
        
        # 2. invalid delete syntax
        with tempfile.NamedTemporaryFile() as user_file:
            with open(user_file.name, 'w') as f:
                f.write("""
                DELETE ALL johndoe%
                """)
            with self.assertLogs(level='ERROR'):
                update_users.update_users_from_file(user_file.name, self.db)
            # make sure both users are in the database
            self.assertEqual(self.db.query(Master).count(), 2)

        # 3. invalid delete syntax
        with tempfile.NamedTemporaryFile() as user_file:
            with open(user_file.name, 'w') as f:
                f.write("""
                DELETEALL johndoe
                """)
            with self.assertLogs(level='ERROR'):
                update_users.update_users_from_file(user_file.name, self.db)
            # make sure both users are in the database
            self.assertEqual(self.db.query(Master).count(), 2)

        # 4. invalid delete syntax
        with tempfile.NamedTemporaryFile() as user_file:
            with open(user_file.name, 'w') as f:
                f.write("""
                DELETE TOO MANY PARAMS johndoe
                """)
            with self.assertLogs(level='ERROR'):
                update_users.update_users_from_file(user_file.name, self.db)
            # make sure both users are in the database
            self.assertEqual(self.db.query(Master).count(), 2)

        # 5. cannot have duplicate API keys
        with tempfile.NamedTemporaryFile() as user_file:
            with open(user_file.name, 'w') as f:
                f.write("""
                newuser, 123abcdefghijkl
                """)
            with self.assertLogs(level='ERROR'):
                update_users.update_users_from_file(user_file.name, self.db)
            # make sure both users are in the database
            self.assertEqual(self.db.query(Master).count(), 2)

        # 6. key too short
        with tempfile.NamedTemporaryFile() as user_file:
            with open(user_file.name, 'w') as f:
                f.write("""
                newuser, abcdef
                """)
            with self.assertLogs(level='ERROR'):
                update_users.update_users_from_file(user_file.name, self.db)
            # make sure both users are in the database
            self.assertEqual(self.db.query(Master).count(), 2)



class ConfigurationError(Exception):
    """
    Error setting up an object due to incorrect configuration
    """
    pass


class SubscriptionError(Exception):
    """
    Error subscribing to a stream that is not available
    """
    pass


class ConnectionError(Exception):
    """
    Error connecting to a Joule server
    """
    pass


class DataError(Exception):
    """
    Error generated by backend storage (timescale)
    """
    pass


class DecimationError(DataError):
    """
    Requested data is not sufficiently decimated
    """
    pass


class ApiError(Exception):
    """
    Error generated by an API call. Catches all API related errors.
    """
    pass

class InvalidDataStreamParameter(ApiError):
    """
    Passing an invalid datatype to an API function that expects a DataStream
    """
    def __init__(self):
        message = "Invalid stream datatype. Must be DataStream, Path, or ID"
        super().__init__(message)

class InvalidEventStreamParameter(ApiError):
    """
    Passing an invalid datatype to an API function that expects an EventStream
    """
    def __init__(self):
        message = "Invalid stream datatype. Must be EventStream, Path, or ID"
        super().__init__(message)

class StreamNotFound(ApiError):
    """
    Type of APIError. Requested stream does not exist
    """
    pass

class PipeError(Exception):
    """
    Error reading from or writing to a pipe
    """
    pass

class EmptyPipeError(PipeError):
    """
    Attempt to read from an empty pipe
    """
    pass



import asyncpg
import json
import datetime
from typing import List, Optional, TYPE_CHECKING, Dict, Callable, Tuple

from joule.models.data_store import psql_helpers
import joule.utilities

if TYPE_CHECKING:
    from joule.models import EventStream


class EventStore:

    def __init__(self, dsn=""):
        self.dsn = dsn
        self.pool = None

    async def initialize(self, pool=None) -> None:
        if pool is None:
            if self.dsn == "":
                raise ValueError("Must specify either dsn or pool")
            self.pool = await asyncpg.create_pool(self.dsn, command_timeout=60)
        else:
            self.pool = pool

    async def create(self, stream: 'EventStream'):
        async with self.pool.acquire() as conn:
            await psql_helpers.create_event_table(conn, stream)

    async def upsert(self, stream: 'EventStream', events: List) -> List:
        updated_event_candidates = list(filter(lambda event: "id" in event and event["id"] is not None, events))
        new_events = list(filter(lambda event: "id" not in event or event["id"] is None, events))
        
        async with self.pool.acquire() as conn:
            # make sure the events with ID's are actually in this table (not copied from somewhere else)
            # If their ID is not found put them in the new_events list instead
            updated_events = []
            for event in updated_event_candidates:
                query = f"SELECT COUNT(*) FROM data.event{stream.id} WHERE id=$1"
                exists = (await conn.fetch(query, event["id"]))[0]
                if exists:
                    updated_events.append(event)
                else:
                    new_events.append(event)

            # Add new events
            if len(new_events) > 0:
                seq = f"'data.event{stream.id}_id_seq'"
                query = await conn.fetch(
                    f"select setval({seq},nextval({seq}) + {len(new_events)} -1) as stop")

                serial = query[0]["stop"] - len(new_events)

                for e in new_events:
                    serial += 1
                    e["id"] = serial
                assert serial == query[0]["stop"]
                await conn.copy_records_to_table(f"event{stream.id}",
                                                 columns=['id',
                                                          'start_time',
                                                          'end_time',
                                                          'content'],
                                                 schema_name='data',
                                                 records=map(event_to_record, new_events))
            # Update existing events
            query = f"UPDATE data.event{stream.id} SET start_time=$2, end_time=$3, content=$4 WHERE id=$1"
            if len(updated_events) > 0:
                mapper = map(event_to_record, updated_events)
                for e in mapper:
                    await conn.execute(query, *e)
        return events

    async def count(self, stream: 'EventStream',
                    start: Optional[int] = None, end: Optional[int] = None,
                    json_filter=None,
                    include_on_going_events=False) -> int:
        if end is not None and start is not None and end <= start:
            raise ValueError("Invalid time bounds start [%d] must be < end [%d]" % (start, end))
        query = f"SELECT count(*) FROM data.event{stream.id} "
        if not include_on_going_events:
            end_col_name = None  # only return events that start within this interval
        else:
            end_col_name = 'end_time'  # return all events that are active within this interval
        where_clauses = [psql_helpers.query_time_bounds(start, end,
                                                        start_col_name='start_time',
                                                        end_col_name=end_col_name)]
        if json_filter is not None and len(json_filter) > 0:
            where_clauses.append(psql_helpers.query_event_json(json_filter))
        where_clauses = [wc for wc in where_clauses if len(wc) > 0]
        if len(where_clauses) > 0:
            query += "WHERE " + " AND ".join(where_clauses)
        async with self.pool.acquire() as conn:
            record = await conn.fetch(query)
            return record[0]['count']
        
    async def histogram(self, stream: 'EventStream',
                        start: Optional[int] = None, end: Optional[int] = None,
                        json_filter=None,
                        nbuckets=100) -> List[Dict]:
        if end is not None and start is not None and end <= start:
            raise ValueError("Invalid time bounds start [%d] must be < end [%d]" % (start, end))
        # if start or end are omitted use the beginning/end of the dataset
        async with self.pool.acquire() as conn:
            if end is None:
                query = f"SELECT end_time FROM data.event{stream.id} ORDER BY end_time DESC LIMIT 1"
                end = (await conn.fetch(query))
                if len(end) == 0:
                    return []  # no data
                end = end[0]['end_time']
            elif type(end) is not datetime.datetime:
                end = datetime.datetime.fromtimestamp(end / 1e6, tz=datetime.timezone.utc).replace(tzinfo=None)
            if start is None:
                query = f"SELECT start_time FROM data.event{stream.id} ORDER BY start_time ASC LIMIT 1"
                start = (await conn.fetch(query))[0]['start_time']
            elif type(start) is not datetime.datetime:
                start = datetime.datetime.fromtimestamp(start / 1e6, tz=datetime.timezone.utc).replace(tzinfo=None)
            bucket_size = ((end - start) / (nbuckets - 1)).total_seconds()
            query = f"""SELECT time_bucket_gapfill('{bucket_size}s', start_time) AS bucket, COALESCE(count(*),0) AS count
                       FROM data.event{stream.id}
                       WHERE start_time >= $1 AND start_time <= $2"""
            if json_filter is not None and len(json_filter) > 0:
                query += " AND " + psql_helpers.query_event_json(json_filter)
            query += " GROUP BY bucket ORDER BY bucket ASC"
            records = await conn.fetch(query, start, end)
            return histogram_to_events(records)

    async def extract(self, stream: 'EventStream',
                      start: Optional[int] = None,
                      end: Optional[int] = None,
                      json_filter=None,
                      limit=None,
                      include_on_going_events=False) -> List[Dict]:
        if end is not None and start is not None and end <= start:
            raise ValueError("Invalid time bounds start [%d] must be < end [%d]" % (start, end))
        query = f"SELECT id, start_time, end_time, content, {stream.id} as event_stream_id FROM data.event{stream.id} "
        if not include_on_going_events:
            end_col_name = None  # only return events that start within this interval
        else:
            end_col_name = 'end_time'  # return all events that are active within this interval
        where_clauses = [psql_helpers.query_time_bounds(start, end,
                                                        start_col_name='start_time',
                                                        end_col_name=end_col_name)]
        if json_filter is not None and len(json_filter) > 0:
            where_clauses.append(psql_helpers.query_event_json(json_filter))
        where_clauses = [wc for wc in where_clauses if len(wc) > 0]
        if len(where_clauses) > 0:
            query += "WHERE " + " AND ".join(where_clauses)
        if limit is not None:
            assert limit > 0, "limit must be > 0"
            if start is None and end is not None:
                query += " ORDER BY start_time DESC"
            else:
                query += " ORDER BY start_time ASC"
            query += f" LIMIT {limit}"
        else:
            query += " ORDER BY start_time ASC"
        async with self.pool.acquire() as conn:
            records = await conn.fetch(query)
            events = list(map(record_to_event, records))
            events.sort(key=lambda e: e["start_time"])
            return events
       
    async def remove(self, stream: 'EventStream', start: Optional[int] = None,
                     end: Optional[int] = None,
                     json_filter=None):
        if start is None and end is None and json_filter is None:
            # flush the stream
            async with self.pool.acquire() as conn:
                return await conn.execute(f"TRUNCATE data.event{stream.id}")
            
        query = f"DELETE FROM data.event{stream.id} "
        where_clauses = [psql_helpers.query_time_bounds(start, end,
                                                        start_col_name='start_time',
                                                        end_col_name='end_time')]
        if json_filter is not None and len(json_filter) > 0:
            where_clauses.append(psql_helpers.query_event_json(json_filter))
        where_clauses = [wc for wc in where_clauses if len(wc) > 0]
        if len(where_clauses) > 0:
            query += " WHERE " + " AND ".join(where_clauses)
        async with self.pool.acquire() as conn:
            return await conn.execute(query)
        
    async def destroy(self, stream: 'EventStream'):
        try:
            async with self.pool.acquire() as conn:
                return await conn.execute(f"DROP TABLE data.event{stream.id}")
        except asyncpg.UndefinedTableError:
            pass  # legacy nodes may not have tables for empty streams

    async def info(self, streams: List['EventStream']) -> Dict[int, 'StreamInfo']:
        if len(streams) == 0:
            return {}  # no event streams
        info_objects = {s.id: StreamInfo(None, None, 0, 0, 0) for s in streams}

        async with self.pool.acquire() as conn:
            for stream in streams:
                query = "select count(*), min(start_time) as start_time," \
                        f" max(end_time) as end_time from data.event{stream.id}"
                row = await conn.fetchrow(query)
                if row['count'] == 0:
                    continue  # no events in stream
                end_time = joule.utilities.datetime_to_timestamp(row['end_time'])
                start_time = joule.utilities.datetime_to_timestamp(row['start_time'])
                total_time = end_time - start_time
                total_bytes = 0  # TODO
                info_objects[stream.id] = StreamInfo(start_time, end_time, row['count'],
                                                        total_time, total_bytes)
        return info_objects

    async def close(self):
        await self.pool.close()


class StreamInfo:
    def __init__(self, start: Optional[int], end: Optional[int], event_count: int,
                 total_time: int = 0, bytes: int = 0):
        self.start = start
        self.end = end
        self.event_count = event_count
        self.bytes = bytes
        self.total_time = total_time

    def __repr__(self):
        return "<StreamInfo start=%r end=%r event_count=%r, total_time=%r>" % (
            self.start, self.end, self.event_count, self.total_time)

    def __eq__(self, other):
        return self.start == other.start and \
               self.end == other.end and \
               self.event_count == other.event_count and \
               self.bytes == other.bytes and \
               self.total_time == other.total_time

    def to_json(self):
        return {
            "start": self.start,
            "end": self.end,
            "event_count": self.event_count,
            "bytes": self.bytes,
            "total_time": self.total_time
        }


def event_to_record(event: Dict) -> Tuple:
    start = joule.utilities.timestamp_to_datetime(event['start_time'])
    if event['end_time'] is not None:
        if event['start_time'] >= event['end_time']:
            raise ValueError("Event end [%d] cannot be before start [%d]" % (
                event['start_time'], event['end_time']
            ))
        end = joule.utilities.timestamp_to_datetime(event['end_time'])
    else:
        end = None
    # return naive datetime objects for storage (expected by asyncpg)
    return event["id"], start.replace(tzinfo=None), end.replace(tzinfo=None), json.dumps(event['content'])


def record_to_event(record: asyncpg.Record) -> Dict:
    if record['end_time'] is not None:
        end = joule.utilities.datetime_to_timestamp(record['end_time'])
    else:
        end = None
    return {
        'id': record['id'],
        'event_stream_id': record['event_stream_id'],
        'start_time': joule.utilities.datetime_to_timestamp(record['start_time']),
        'end_time': end,
        'content': json.loads(record['content'])}


def histogram_to_events(histogram: List[Tuple[int, int]]) -> List[Dict]:
    events = []
    for i in range(len(histogram) - 1):
        events.append({
            "id": -1,  # not used
            "start_time": joule.utilities.datetime_to_timestamp(histogram[i][0]),
            "end_time": joule.utilities.datetime_to_timestamp(histogram[i + 1][0]),
            "content": {"count": histogram[i][1]}})
    events.append({
        "id": -1,  # not used
        "start_time": joule.utilities.datetime_to_timestamp(histogram[-1][0]),
        "end_time": joule.utilities.datetime_to_timestamp(histogram[-1][0]),
        "content": {"count": histogram[-1][1]}})
    return events



from aiohttp.test_utils import AioHTTPTestCase, unittest_run_loop
from aiohttp import web
from sqlalchemy.orm import Session
import json
import asyncio
from joule.models import folder, DataStream, Folder
import joule.controllers
from tests.controllers.helpers import create_db, MockStore, MockEventStore
from joule import app_keys
from joule.constants import EndPoints, ApiErrorMessages

import testing.postgresql
psql_key = web.AppKey("psql", testing.postgresql.Postgresql)

class TestFolderControllerErrors(AioHTTPTestCase):

    async def tearDownAsync(self):
        self.app[app_keys.db].close()
        self.app[psql_key].stop()
        await self.client.close()

    async def get_application(self):
        app = web.Application()
        app.add_routes(joule.controllers.routes)
        # this takes a while, adjust the expected coroutine execution time
        loop = asyncio.get_running_loop()
        loop.slow_callback_duration = 2.0
        app[app_keys.db], app[psql_key] = create_db(["/top/leaf/stream1:float32[x, y, z]",
                                            "/top/same_name/stream3:float32[x, y, z]",
                                            "/top/middle/leaf/stream2:int16[val1, val2]"])
        app[app_keys.data_store] = MockStore()
        app[app_keys.event_store] = MockEventStore()
        return app


    async def test_folder_info(self):
        # must specify an id or path
        resp = await self.client.get(EndPoints.folder, params={})
        self.assertEqual(resp.status, 400)
        self.assertIn(ApiErrorMessages.specify_id_or_path, await resp.text())

        # folder must exist
        resp = await self.client.get(EndPoints.folder, params={"id": 2343})
        self.assertEqual(resp.status, 404)
        self.assertIn("does not exist", await resp.text())


    async def test_folder_move(self):
        db: Session = self.app[app_keys.db]

        # must be a json request
        resp = await self.client.put(EndPoints.folder_move, data={"dest_path": "/new/location"})
        self.assertEqual(resp.status, 400)
        self.assertIn("json", await resp.text())

        # must specify an id or a path
        resp = await self.client.put(EndPoints.folder_move, json={"dest_path": "/new/location"})
        self.assertEqual(resp.status, 400)
        # must specify a destination
        resp = await self.client.put(EndPoints.folder_move, json={"src_path": "/top/leaf"})
        self.assertEqual(resp.status, 400)
        # return "not found" on bad id
        resp = await self.client.put(EndPoints.folder_move, json={"src_id": 2348,
                                                                "dest_path": "/x/y"})
        self.assertEqual(resp.status, 404)
        # return "not found" on bad path
        resp = await self.client.put(EndPoints.folder_move, json={"src_path": "/missing/folder",
                                                                "dest_patbh": "/x/y"})
        self.assertEqual(resp.status, 404)
        # error on bad destination
        resp = await self.client.put(EndPoints.folder_move, json={"src_path": "/top/leaf",
                                                                "dest_path": "malformed"})
        self.assertEqual(resp.status, 400)

        # cannot conflict with an existing folder in the destination
        resp = await self.client.put(EndPoints.folder_move, json={"src_path": "/top/leaf",
                                                                "dest_path": "/top/middle"})
        self.assertEqual(resp.status, 400)

        # cannot move a folder into its children
        resp = await self.client.put(EndPoints.folder_move, json={"src_path": "/top",
                                                                "dest_path": "/top/leaf"})
        self.assertEqual(resp.status, 400)
        self.assertTrue('parent' in await resp.text())
        self.assertIsNotNone(folder.find('/top/middle/leaf', db))

        # cannot move folders with locked streams
        my_stream: DataStream = db.query(DataStream).filter_by(name="stream1").one()
        my_stream.is_configured = True
        resp = await self.client.put(EndPoints.folder_move, json={"src_path": "/top/leaf",
                                                                "dest_path": "/top/middle"})
        self.assertEqual(resp.status, 400)
        self.assertTrue('locked' in await resp.text())


    async def test_folder_delete(self):
        db: Session = self.app[app_keys.db]

        resp = await self.client.delete(EndPoints.folder, params={})
        self.assertEqual(resp.status, 400)
        # return "not found" on bad id
        resp = await self.client.delete(EndPoints.folder, params={"id": 2348})
        self.assertEqual(resp.status, 404)
        # return "not found" on bad path
        resp = await self.client.delete(EndPoints.folder, params={"path": "/bad/path"})
        self.assertEqual(resp.status, 404)
        # cannot delete folders with children unless recursive is set
        f = folder.find("/a/new/path", db, create=True)
        resp = await self.client.delete(EndPoints.folder, params={"path": "/a"})
        self.assertEqual(resp.status, 400)
        self.assertTrue('recursive' in await resp.text())
        self.assertIsNotNone(db.get(Folder, f.id))
        # cannot delete locked folders
        my_folder = folder.find("/top/middle", db)
        my_folder.children[0].data_streams[0].is_configured = True
        resp = await self.client.delete(EndPoints.folder, params={"path": "/top/middle"})
        self.assertEqual(resp.status, 400)
        self.assertIn("locked", await resp.text())



    async def test_folder_update(self):
        db: Session = self.app[app_keys.db]
        my_folder = folder.find("/top/middle", db)

        # must be a json request
        resp = await self.client.put(EndPoints.folder, data={"bad": "notjson"})
        self.assertEqual(resp.status, 400)
        self.assertIn("json", await resp.text())

        # invalid name, nothing should be saved
        payload = {
            "id": my_folder.id,
            "folder":
                {"name": "",
                 "description": "new description"}
        }
        resp = await self.client.put(EndPoints.folder, json=payload)
        self.assertEqual(resp.status, 400)
        self.assertTrue('name' in await resp.text())
        my_folder = folder.find("/top/middle", db)
        self.assertEqual("middle", my_folder.name)

        # request must specify an id
        payload = {
            "folder": json.dumps({"name": "new name"})
        }
        resp = await self.client.put(EndPoints.folder, json=payload)
        self.assertEqual(resp.status, 400)
        self.assertTrue('id' in await resp.text())

        # request must have folder attr
        payload = {
            "id": my_folder.id
        }
        resp = await self.client.put(EndPoints.folder, json=payload)
        self.assertEqual(resp.status, 400)
        self.assertTrue('folder' in await resp.text())

        # folder attr must be valid json
        payload = {
            "id": my_folder.id,
            "folder": "notjson"
        }
        resp = await self.client.put(EndPoints.folder, json=payload)
        self.assertEqual(resp.status, 400)
        self.assertTrue('JSON' in await resp.text())

        # cannot modify locked streams
        my_folder.children[0].data_streams[0].is_configured = True
        payload = {
            "id": my_folder.id,
            "folder": json.dumps({"name": "new name"})
        }
        resp = await self.client.put(EndPoints.folder, json=payload)
        self.assertEqual(resp.status, 400)
        self.assertTrue('locked' in await resp.text())
        my_folder.children[0].data_streams[0].is_configured = False

        # folder must exist
        payload = {
            "id": 4898,
            "stream": json.dumps({"name": "new name"})
        }
        resp = await self.client.put(EndPoints.folder, json=payload)
        self.assertEqual(resp.status, 404)
        self.assertTrue('exist' in await resp.text())

        # folder name cannot conflict with a peer
        payload = {
            "id": my_folder.id,
            "folder":
                {"name": "same_name",
                 "description": "new description"}
        }
        resp = await self.client.put(EndPoints.folder, json=payload)
        self.assertEqual(resp.status, 400)
        self.assertIn('folder with this name', await resp.text())
        my_folder = folder.find("/top/middle", db)
        self.assertEqual("middle", my_folder.name)



from click.testing import CliRunner
import os
import logging
import csv
import warnings
from tests.cli.fake_joule import FakeJoule, FakeJouleTestCase
from joule.cli import main

ANNOTATION_LIST = os.path.join(os.path.dirname(__file__), 'annotations.json')
warnings.simplefilter('always')
aio_log = logging.getLogger('aiohttp.access')
aio_log.setLevel(logging.WARNING)


class TestStreamAnnotation(FakeJouleTestCase):

    def test_lists_annotations_csv(self):
        server = FakeJoule()
        with open(ANNOTATION_LIST, 'r') as f:
            server.response = f.read()
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['stream', 'annotations', '/my/stream', '--csv'])
        self._assert_no_error(result)
        output = result.output
        lines = output.split("\n")
        reader = csv.reader(lines[1:])
        annotations = []
        for row in reader:
            annotations.append(row)
        # first annotation is an event
        self.assertEqual(['Instant', 'instant temperature', '1561752347819968'],
                         annotations[0])
        # second annoation is a range with no comment
        self.assertEqual(['hot!!!', '', '1561861116610000', '1561862838572000'],
                         annotations[1])
        self.stop_server()

    def test_lists_all_annotations_table(self):
        server = FakeJoule()
        with open(ANNOTATION_LIST, 'r') as f:
            server.response = f.read()
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['stream', 'annotations', '/my/stream'])
        self.assertEqual(result.exit_code, 0)
        output = result.output
        # check for a few lines in the table
        self.assertIn("Instant", output)
        # timestamps should be displayed as dates
        self.assertNotIn("1561753681898272", output)
        self.assertIn("Fri, 28 Jun 2019 16:05:47", output)
        self.stop_server()

    def test_reteives_time_bounded_annotations(self):
        server = FakeJoule()
        with open(ANNOTATION_LIST, 'r') as f:
            server.response = f.read()
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['stream', 'annotations', '/my/stream',
                                      '--start', '28 Jun 2019 16:00',
                                      '--end', '29 Jun 2019 16:00'])
        self.assertEqual(result.exit_code, 0)
        query_params = self.msgs.get()
        self.assertEqual(query_params["start"], "1561752000000000")
        self.assertEqual(query_params["end"], "1561838400000000")
        self.stop_server()

    def test_delete_all_annotations(self):
        server = FakeJoule()
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['stream', 'annotations', '/my/stream', '--delete'])
        self._assert_no_error(result)
        query_params = self.msgs.get()
        self.assertEqual(query_params["stream_path"], "/my/stream")
        self.stop_server()

    def test_delete_time_bounded_annotations(self):
        server = FakeJoule()
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['stream', 'annotations', '/my/stream',
                                      '--delete',
                                      '--start', '28 Jun 2019 16:00 EDT',
                                      '--end', '29 Jun 2019 16:00 EDT'])
        self._assert_no_error(result)
        query_params = self.msgs.get()
        self.assertEqual(query_params["stream_path"], "/my/stream")
        self.assertEqual(query_params["start"], "1561752000000000")
        self.assertEqual(query_params["end"], "1561838400000000")
        self.stop_server()

    def test_start_time_bound_errors(self):
        server = FakeJoule()
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['stream', 'annotations', '/my/stream',
                                      '--start', 'baddate',
                                      '--end', '29 Jun 2019 16:00'])
        self.assertIn('start time', result.output)
        self.assertEqual(result.exit_code, 1)
        self.stop_server()

    def test_end_time_bound_errors(self):
        server = FakeJoule()
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['stream', 'annotations', '/my/stream',
                                      '--start', '28 Jun 2019 16:00',
                                      '--end', 'baddate'])
        self.assertIn('end time', result.output)
        self.assertEqual(result.exit_code, 1)
        self.stop_server()

    def test_when_server_returns_error_code(self):
        server = FakeJoule()
        server.response = "test error"
        server.http_code = 500
        server.stub_annotation = True
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['stream', 'annotations', 'folder/stream'])
        self.assertIn('500', result.output)
        self.assertIn("test error", result.output)
        self.assertEqual(result.exit_code, 1)
        self.stop_server()

    def _assert_no_error(self, result):
        if result.exit_code != 0:
            print("ERROR: ", result.output)
            print("EXCEPTION", result.exception)
        self.assertEqual(result.exit_code, 0)



import os
import argparse
import asyncio
import json
import numpy as np

from .test_filter_module import SimpleFilter
from .test_reader_module import SimpleReader
from joule.client import CompositeModule
from joule.api import DataStream
from joule.models import DataStream, Element, pipes
from tests import helpers
import warnings

warnings.simplefilter('always')


class SimpleComposite(CompositeModule):
    # runs the simple reader and simple filter
    # as a composite module
    async def setup(self, parsed_args, inputs, outputs):
        reader_module = SimpleReader()
        filter_module = SimpleFilter()

        pipe = pipes.LocalPipe("float32_3")
        reader_task = reader_module.run(parsed_args, pipe)
        filter_task = filter_module.run(parsed_args, {'to_filter': pipe},
                                        {'from_filter': outputs['output']})
        return [reader_task, filter_task]


class TestCompositeModule(helpers.AsyncTestCase):

    def setUp(self):
        super().setUp()
        # module output is a float32_3 stream
        self.stream = DataStream(name="output", datatype=DataStream.DATATYPE.FLOAT32,
                                 elements=[Element(name="e%d" % j, index=j,
                                               display_type=Element.DISPLAYTYPE.CONTINUOUS) for j in range(3)])

    def test_runs_composited_modules(self):
        module = SimpleComposite()
        (r, w) = os.pipe()
        rf = pipes.reader_factory(r)
        pipe = pipes.InputPipe(name="output", stream=self.stream, reader_factory=rf)
        pipe_arg = json.dumps(json.dumps({"outputs": {'output': {'fd': w, 'id': None, 'layout': self.stream.layout}},
                                          "inputs": {}}))
        data = helpers.create_data(self.stream.layout)
        args = argparse.Namespace(pipes=pipe_arg, socket="unset",
                                  node="", api_socket="",
                                  url='http://localhost:8080',
                                  mock_data=data)
        # run the composite module
        module.start(args)
        asyncio.set_event_loop(self.loop)
        # check the output
        received_data = asyncio.run(pipe.read())
        np.testing.assert_array_equal(data['timestamp'], received_data['timestamp'])
        np.testing.assert_array_almost_equal(data['data'] * 2, received_data['data'])
        asyncio.run(pipe.close())

    def test_handles_bad_pipe_configs(self):
        args = argparse.Namespace(pipes="invalid config", socket="unset",
                                  node="", api_socket="",
                                  url='http://localhost:8080')
        module = SimpleComposite()
        with self.assertLogs(level="ERROR"):
            module.start(args)
        asyncio.set_event_loop(self.loop)




from click.testing import CliRunner
from tests.cli.fake_joule import FakeJouleTestCase
from joule.cli import main
from ..fake_joule import print_result_on_error
import unittest
import tempfile
import shutil
import os
from joule.errors import ApiError
from unittest.mock import patch
from joule.api import helpers, TcpNode

CAFILE = os.path.join(os.path.dirname(__file__), 'ca.joule.crt')

class TestNodeList(unittest.TestCase):

    def setUp(self):
        self.temp_dir = tempfile.TemporaryDirectory()
        os.environ["JOULE_USER_CONFIG_DIR"] = self.temp_dir.name
        self.ca_path = os.path.join(self.temp_dir.name, "ca.crt")
        shutil.copyfile(CAFILE, self.ca_path)
        
    def tearDown(self):
        self.temp_dir.cleanup()

    def test_lists_nodes(self):
        node1 = TcpNode("node1", "https://localhost:8088",
                        "node1_key")
        node2 = TcpNode("node2", "https://localhost:8089",
                        "node2_key")

        helpers.save_node(node1)
        helpers.save_node(node2)
        runner = CliRunner()
        result = runner.invoke(main, ['node', 'list'])
        print_result_on_error(result)
        self.assertEqual(result.exit_code, 0)
        output = result.output
        # make sure nodes are listed
        for name in ['node1', 'node2']:
            self.assertIn(name, output)

    @patch('joule.cli.node.list.get_node')
    @patch('joule.cli.node.list.get_nodes')
    def test_handles_api_error(self, get_nodes, get_node):
        get_node.side_effect = ApiError("test error")
        runner = CliRunner()
        result = runner.invoke(main, ['node', 'list'])
        self.assertIn('test error', result.output)
        self.assertEqual(result.exit_code, 1)
        



import click
import datetime
import asyncio

from joule import errors
from joule.cli.config import pass_config


@click.command(name="info")
@click.argument("path")
@pass_config
def cli_info(config, path):
    """Display event stream information."""
    try:
        asyncio.run(
            _run(config.node, path))
    except errors.ApiError as e:
        raise click.ClickException(str(e)) from e
    finally:
        asyncio.run(
            config.close_node())


async def _run(node, path):
    my_stream = await node.event_stream_get(path)
    my_info = await node.event_stream_info(my_stream)
    # display stream information
    click.echo()
    click.echo("Event Stream Information:")
    click.echo("\tName:         %s" % my_stream.name)
    click.echo("\tStart:        %s" % _display_time(my_info.start))
    click.echo("\tEnd:          %s" % _display_time(my_info.end))
    click.echo("\tEvents:       %d" % my_info.event_count)
    click.echo("")
    if len(my_stream.event_fields) > 0:
        click.echo("Event Fields:")
        click.echo("\tType\t| Name")
        click.echo("\t--------------------")
        for name, datatype in my_stream.event_fields.items():
            click.echo(f"\t{datatype}\t| {name}")
    else:
        click.echo("There are no fields specified for this event stream")

def _display_time(time: int) -> str:
    if time is None:
        return u"\u2014"  # emdash
    return str(datetime.datetime.fromtimestamp(time / 1e6))



from click.testing import CliRunner
import os
import warnings
import re

from ..fake_joule import FakeJoule, FakeJouleTestCase
from joule.cli import main

EVENT_INFO = os.path.join(os.path.dirname(__file__), 'event_info.json')
EMPTY_EVENT_INFO = os.path.join(os.path.dirname(__file__), 'empty_event_info.json')

warnings.simplefilter('always')


class TestEventCopy(FakeJouleTestCase):

    def test_copies_all_data_to_new_stream(self):
        pass

    def test_copies_region_of_data(self):
        pass

    def test_replaces_existing_data_on_copy(self):
        pass

    def test_ignores_existing_data_on_copy(self):
        pass

    def test_errors_on_invalid_stream(self):
        pass

    def test_errors_when_source_is_empty(self):
        pass


import numpy as np
import io
from struct import pack
import datetime
import asyncpg
from typing import List, Optional, Tuple, Dict
import logging
from joule.utilities import datetime_to_timestamp as dt2ts
import psycopg2.sql

from joule.errors import DataError
from joule.models.data_stream import DataStream
from joule.models.event_stream import EventStream
import joule.utilities

log = logging.getLogger('joule')

postgres_ts_offset = 946684800000000  # January 1 2000 GMT


def data_to_bytes(data: np.ndarray) -> io.BytesIO:
    pgcopy_dtype = [("num_fields", ">i2"),
                    ("time_length", '>i4'),
                    ("time", '>i8')]
    dtype_tuple = data.dtype.descr[1]
    elem_dtype = dtype_tuple[1].replace('<', '>')
    elem_length = data['data'].dtype.alignment
    if len(dtype_tuple) == 3:
        n_elem = dtype_tuple[2][0]
    else:
        n_elem = 1
    for i in range(n_elem):
        pgcopy_dtype += [("elem%d_length" % i, '>i4'),
                         ("elem%d" % i, elem_dtype)]
    pgcopy = np.empty(data.shape, pgcopy_dtype)
    pgcopy['num_fields'] = n_elem + 1
    pgcopy['time_length'] = 8
    pgcopy['time'] = data['timestamp'] - postgres_ts_offset
    for i in range(n_elem):
        pgcopy['elem%d_length' % i] = elem_length
        pgcopy['elem%d' % i] = data['data'][:, i]
    cpy = io.BytesIO()
    # signature, flag field and and header extension (both empty)
    cpy.write(pack('!11sii', b'PGCOPY\n\377\r\n\0', 0, 0))
    cpy.write(pgcopy.tobytes())
    cpy.write(pack('!h', -1))
    cpy.seek(0)
    return cpy


def bytes_to_data(buffer: io.BytesIO, dtype: np.dtype) -> np.ndarray:
    pgcopy_dtype = [("num_fields", ">i2"),
                    ("time_length", '>i4'),
                    ("time", '>i8')]
    dtype_tuple = dtype.descr[1]
    elem_dtype = dtype_tuple[1].replace('<', '>')
    elem_length = dtype['data'].alignment
    if len(dtype_tuple) == 3:
        n_elem = dtype_tuple[2][0]
    else:
        n_elem = 1
    for i in range(n_elem):
        pgcopy_dtype += [("elem%d_length" % i, '>i4'),
                         ("elem%d" % i, elem_dtype)]
    pgcopy_dtype = np.dtype(pgcopy_dtype)
    nbytes = buffer.seek(0, io.SEEK_END)
    buffer.seek(0)
    # check the header
    header = pack('!11sii', b'PGCOPY\n\377\r\n\0', 0, 0)
    rx_header = buffer.read(len(header))
    if rx_header != header:
        raise DataError("bad pgcopy header")
    row_size = pgcopy_dtype.itemsize
    if (nbytes - 21) % row_size != 0:
        raise DataError("invalid number of data bytes")
    nrows = (nbytes - 21) // row_size
    tuple_data = np.frombuffer(buffer.read(nbytes - 21), pgcopy_dtype)
    rx_data = np.empty(nrows, dtype)
    rx_data['timestamp'] = tuple_data['time'] + postgres_ts_offset
    for i in range(n_elem):
        rx_data['data'][:, i] = tuple_data['elem%d' % i]
    # reader the footer
    footer = pack('!h', -1)
    rx_footer = buffer.read()
    if footer != rx_footer:
        raise DataError("ERROR: invalid footer")
    return rx_data


def query_time_bounds(start, end, start_col_name='time', end_col_name=None):
    # bounds are [ --- )
    limits = []
    if start is not None:
        if type(start) is not datetime.datetime:
            start = datetime.datetime.fromtimestamp(start / 1e6, tz=datetime.timezone.utc)
        if end_col_name is not None:  # interval query
            limits.append(f"{end_col_name} >= '{start}'")
        else:  # point query
            limits.append(f"{start_col_name} >= '{start}'")
    if end is not None:
        if type(end) is not datetime.datetime:
            end = datetime.datetime.fromtimestamp(end / 1e6, tz=datetime.timezone.utc)
        limits.append(f"{start_col_name} < '{end}'")
    if len(limits) > 0:
        return ' AND '.join(limits)
    else:
        return ''


EventFilter = Tuple[str, str, str]
EventFilterGroup = List[EventFilter]  # AND clauses
EventFilterGroups = List[EventFilterGroup]  # OR between groups


def query_event_json(filter_groups: EventFilter) -> str:
    groups = []
    for group in filter_groups:
        clauses = []
        for clause in group:
            if len(clause) != 3:
                raise ValueError("invalid clause, must be [key,comparison,value]")
            key, comparison, value = clause
            key_safe = key.replace("'", "''")
            if comparison in ['is', 'not', 'like', 'unlike']:
                value_safe = value.replace("'", "''")
                if comparison == 'is':
                    clauses.append(f"(content->>'{key_safe}'='{value_safe}')")
                elif comparison == 'not':
                    clauses.append(f"(NOT (content->>'{key_safe}'='{value_safe}'))")
                elif comparison == 'like':
                    clauses.append(f"(content->>'{key_safe}' like '{value_safe}')")
                elif comparison == 'unlike':
                    clauses.append(f"(content->>'{key_safe}' not like '{value_safe}')")
            elif comparison == 'gt':
                clauses.append(f"(CAST (content->>'{key_safe}' AS FLOAT)>{float(value)})")
            elif comparison == 'gte':
                clauses.append(f"(CAST (content->>'{key_safe}' AS FLOAT)>={float(value)})")
            elif comparison == 'lt':
                clauses.append(f"(CAST (content->>'{key_safe}' AS FLOAT)<{float(value)})")
            elif comparison == 'lte':
                clauses.append(f"(CAST (content->>'{key_safe}' AS FLOAT)<={float(value)})")
            elif comparison == 'eq':
                clauses.append(f"(CAST (content->>'{key_safe}' AS FLOAT)={float(value)})")
            elif comparison == 'neq':
                clauses.append(f"(NOT CAST (content->>'{key_safe}' AS FLOAT)={float(value)})")
            else:
                raise ValueError(f"Invalid comparison {comparison}, must be [gt|gte|lt|lte|eq|neq|is|not|like|unlike]")
        groups.append(' AND '.join(clauses))
    sql_query = " OR ".join(groups)
    return f"({sql_query})"


async def create_event_table(conn: asyncpg.Connection, stream: 'EventStream'):
    # first check if the table exists
    table_name = f"data.event{stream.id}"
    sql = f"SELECT to_regclass('{table_name }');"
    result = await conn.fetchval(sql)
    if result is not None:
        return  # nothing to do, table exists

    sql = f"""CREATE TABLE {table_name } (
    id SERIAL,
    start_time TIMESTAMP NOT NULL,
    end_time TIMESTAMP NOT NULL,
    content JSONB
    )"""
    await conn.execute(sql)
    if stream.chunk_duration_us > 0:
        sql = f"""SELECT create_hypertable('{table_name }', 'start_time')"""
        await conn.execute(sql)
        await update_chunk_interval(conn, table_name, stream.chunk_duration_us)
    sql = f"""CREATE INDEX ON {table_name} (start_time, end_time)"""
    await conn.execute(sql)
    sql = f"""GRANT SELECT ON {table_name} TO joule_module"""
    await conn.execute(sql)


async def create_stream_table(conn: asyncpg.Connection, stream: DataStream):
    n_elems = len(stream.elements)
    # create the main table
    col_type = get_psql_type(stream.datatype)
    cols = ["elem%d %s NOT NULL" % (x, col_type) for x in range(n_elems)]
    sql = "CREATE TABLE IF NOT EXISTS data.stream%d (" % stream.id + \
          "time TIMESTAMP NOT NULL," + \
          ', '.join(cols) + \
          ", PRIMARY KEY(time));"
    await conn.execute(sql)
    sql = f"SELECT create_hypertable('data.stream{stream.id}', 'time', if_not_exists=>true);"
    await conn.execute(sql)
    sql = f"GRANT SELECT ON data.stream{stream.id} TO joule_module;"
    await conn.execute(sql)
    # create interval table
    sql = f"CREATE TABLE IF NOT EXISTS data.stream{stream.id}_intervals (time TIMESTAMP NOT NULL);"
    await conn.execute(sql)


async def create_decimation_table(conn: asyncpg.Connection, stream: DataStream, level: int):
    n_elems = len(stream.elements)
    table_name = 'data.stream%d_%d' % (stream.id, level)
    # create decimation table (just a template)
    mean_cols = ["elem%d REAL NOT NULL" % x for x in range(n_elems)]
    min_cols = ["elem%d_min REAL NOT NULL" % x for x in range(n_elems)]
    max_cols = ["elem%d_max REAL NOT NULL" % x for x in range(n_elems)]
    cols = mean_cols + min_cols + max_cols
    sql = "CREATE TABLE IF NOT EXISTS %s (" % table_name + \
          "time TIMESTAMP NOT NULL," + \
          ', '.join(cols) + \
          ", PRIMARY KEY(time));"
    await conn.execute(sql)
    # use the default chunk interval, this will be updated by the decimator process
    sql = f"""
        SELECT create_hypertable('{table_name}', 'time',
        if_not_exists=>true);
    """
    # print(f"SQL: {sql}")
    await conn.execute(sql)


async def drop_decimation_tables(conn: asyncpg.Connection, stream: DataStream):
    tables = await get_decimation_table_names(conn, stream, with_schema=True)
    for table in tables:
        try:
            await conn.execute(f"DROP TABLE {table}")
        except asyncpg.UndefinedTableError:
            pass


def get_psql_type(x: DataStream.DATATYPE):
    if x == DataStream.DATATYPE.FLOAT32:
        return 'real'
    elif x == DataStream.DATATYPE.FLOAT64:
        return 'double precision'
    elif x == DataStream.DATATYPE.INT16:
        return 'smallint'
    elif x == DataStream.DATATYPE.INT32:
        return 'integer'
    elif x == DataStream.DATATYPE.INT64:
        return 'bigint'
    else:
        # NOTE: All other datatypes have been removed from DATATYPE
        # so this code is unreachable, but it is left here to catch
        # errors from any future changes to DATATYPE
        raise DataError("Invalid type [%r] for timescale backend" % x)


async def get_row_count(conn: asyncpg.Connection, stream: DataStream,
                        start=None, end=None):
    # hyper table approximate row count is not sensitive to data removal
    # always use the custom function
    try:
        bounds = await convert_time_bounds(conn, stream, start, end)
    except asyncpg.UndefinedTableError:
        return 0  # no data tables for this stream
    if bounds is None:
        return 0  # no data
    start, end = bounds
    query = "SELECT stream_row_count(%d, '%s', '%s')" % (stream.id, start, end)
    try:
        nrows = await conn.fetchval(query)
    except asyncpg.UndefinedTableError:
        return 0  # no data tables for this stream
    return nrows


async def get_closest_ts(conn: asyncpg.Connection, stream: DataStream, ts: int) -> Optional[datetime.datetime]:
    base_table = "data.stream%d" % stream.id
    dt = datetime.datetime.fromtimestamp(int(ts) / 1e6, tz=datetime.timezone.utc)
    query = f"SELECT time FROM {base_table} WHERE time < '{dt}' ORDER BY time DESC LIMIT 1"
    #print(f"stream: {stream.id} ts: {ts}: query: {query}")

    try:
        last_dt = await conn.fetchval(query)
    except asyncpg.UndefinedTableError:
        # no data tables so no previous timestamp
        return None
    if last_dt is None:
        # no data exists before ts so no previous timestamp
        return None
    # convert the datetime.datetime to a timestamp
    return dt2ts(last_dt)
    
async def remove_interval_breaks(conn: asyncpg.Connection, stream: DataStream, start: int, end: int):
    # remove interval breaks that are no longer needed
    interval_table = "data.stream%d_intervals" % stream.id
    start = datetime.datetime.fromtimestamp(start / 1e6, tz=datetime.timezone.utc)
    end = datetime.datetime.fromtimestamp(end / 1e6, tz=datetime.timezone.utc)
    query = f"DELETE FROM {interval_table} WHERE time >= '{start}' AND time <= '{end}'"
    try:
        await conn.fetch(query)
    except asyncpg.UndefinedTableError:
        return  # no interval table so no breaks to remove

async def close_interval(conn: asyncpg.Connection, stream: DataStream, ts: int):
    # place a boundary 1us *after* ts
    base_table = "data.stream%d" % stream.id
    interval_table = "data.stream%d_intervals" % stream.id
    ts = datetime.datetime.fromtimestamp(ts / 1e6, tz=datetime.timezone.utc)
    # find the most recent data before this boundary (ts)
    query = "SELECT time FROM %s WHERE time <= '%s' ORDER BY time DESC LIMIT 1" % (base_table, ts)
    try:
        last_ts = await conn.fetchval(query)
    except asyncpg.UndefinedTableError:
        # no data tables so no need for an interval boundary
        return
    if last_ts is None:
        # no data exists before ts so no need for an interval boundary
        return
    # check if this interval is necessary
    query = "SELECT time FROM %s WHERE time <= '%s' ORDER BY time DESC LIMIT 1" % (interval_table, ts)
    last_interval = await conn.fetchval(query)
    if last_interval is None or last_ts > last_interval:
        query = "INSERT INTO %s(time) VALUES ($1)" % interval_table
        await conn.execute(query, last_ts + datetime.timedelta(microseconds=1))


async def get_decimation_table_names(conn: asyncpg.Connection, stream: DataStream, with_schema=True) -> List[str]:
    query = r'''select table_name from information_schema.tables 
                  where table_schema='data' 
                  and table_type='BASE TABLE' 
                  and table_name like 'stream%d\_%%';''' % stream.id
    records = await conn.fetch(query)
    all_tables = [r['table_name'] for r in records]
    # omit the interval table
    data_tables = [name for name in all_tables if not name.endswith('intervals')]
    # print(f"data tables: {data_tables}")
    if with_schema:
        return [f'data.{name}' for name in data_tables]
    else:
        return data_tables


async def get_table_names(conn: asyncpg.Connection, stream: DataStream, with_schema=True) -> List[str]:
    tables = await get_decimation_table_names(conn, stream, with_schema)
    if with_schema:
        return tables + [f'data.stream{stream.id}', f'data.stream{stream.id}_intervals']
    else:
        return tables + [f'stream{stream.id}', f'stream{stream.id}_intervals']


async def get_all_table_names(conn: asyncpg.Connection, with_schema=True) -> List[str]:
    """Return a list of all data tables, used when dropping all data from the node"""
    query = r'''select table_name from information_schema.tables 
               where table_schema='data' 
               and table_type='BASE TABLE' 
               and table_name like 'stream%';'''
    records = await conn.fetch(query)
    if with_schema:
        return ['data.' + r['table_name'] for r in records]
    else:
        return [r['table_name'] for r in records]


async def get_boundaries(conn: asyncpg.Connection, stream: DataStream,
                         start: Optional[int], end: Optional[int]) -> List[datetime.datetime]:
    """
    Return a list of data boundaries including the start and end of the data
    """
    bounds = await convert_time_bounds(conn, stream, start, end)
    if bounds is None:
        return []  # no data so no boundaries
    start, end = bounds
    query = "SELECT time FROM data.stream%d_intervals " % stream.id
    where_clause = query_time_bounds(start, end)
    if len(where_clause)>0:
        query += " WHERE " + where_clause
    query += " ORDER BY time ASC"
    try:
        records = await conn.fetch(query)
    except asyncpg.UndefinedTableError:
        return []  # no data tables so no boundaries
    ts = [start] + [r['time'] for r in records] + [end]
    return [x.replace(tzinfo=datetime.timezone.utc) for x in ts]


async def convert_time_bounds(conn: asyncpg.Connection,
                              stream: DataStream,
                              start: Optional[int], end: Optional[int]) -> Optional[
    Tuple[datetime.datetime, datetime.datetime]]:
    """
    Convert Unix us timestamps to datetime objects and populate [None] values with the
    start or end of the data respectively
    """
    if start is None:
        query = "SELECT time FROM data.stream%d ORDER BY time ASC LIMIT 1" % stream.id
        try:
            start = await conn.fetchval(query)
        except asyncpg.UndefinedTableError:
            return None  # no data tables so no valid time bounds
        if start is None:
            # remove intervals?
            return None  # no data so no valid time bounds
        start = start.replace(tzinfo=datetime.timezone.utc)

    else:
        start = datetime.datetime.fromtimestamp(start / 1e6, tz=datetime.timezone.utc)
    if end is None:
        query = "SELECT time FROM data.stream%d ORDER BY time DESC LIMIT 1" % stream.id
        try:
            end = await conn.fetchval(query) + datetime.timedelta(microseconds=1)
        except asyncpg.UndefinedTableError:
            return None  # no data tables so no valid time bounds
        if end is None:
            # remove intervals?
            return None  # no data so no valid time bounds
        end = end.replace(tzinfo=datetime.timezone.utc)
    else:
        end = datetime.datetime.fromtimestamp(end / 1e6, tz=datetime.timezone.utc)
    return start, end


async def limit_time_bounds(conn: asyncpg.Connection,
                            stream: DataStream,
                            start: Optional[int], end: Optional[int]) -> Tuple[Optional[int], Optional[int]]:
    data_bounds = await convert_time_bounds(conn, stream, None, None)
    if data_bounds is None:
        return None, None
    data_start = joule.utilities.datetime_to_timestamp(data_bounds[0])
    data_end = joule.utilities.datetime_to_timestamp(data_bounds[1])
    if start is None:
        start = data_start
    else:
        start = max(start, data_start)
    if end is None:
        end = data_end
    else:
        end = min(end, data_end)
    return start, end


async def update_chunk_interval(conn: asyncpg.Connection, table: str,
                                chunk_interval_us: int):
    if chunk_interval_us == 0:
        return  # invalid chunk interval setting, ignore
    query = f"SELECT set_chunk_time_interval('{table}', {round(chunk_interval_us)})"
    # print(f"QUERY: <{query}>, {round(chunk_interval_ms/(1000*60*60))}") # show in hours
    await conn.execute(query)



from typing import Optional, Dict, Union, List
import json
import re
import logging
log = logging.getLogger('joule')

from .session import BaseSession
from .folder_type import Folder
from .event import Event
from .event import from_json as event_from_json
from joule import errors
from joule.utilities.validators import validate_event_fields
from joule.constants import EndPoints

EVENT_READ_BLOCK_SIZE = 1000 # Read 1000 events at a time

class EventStream:
    """
        API EventStream model. See :ref:`sec-node-event-stream-actions` for details on using the API to
        manipulate event streams.

        Parameters:
            name (str): stream name, must be unique in the parent
            description (str): optional field
    """

    def __init__(self, name: str = "",
                 description: str = "",
                 chunk_duration: str = "",
                 chunk_duration_us: Optional[int] = None,
                 keep_us: int = -1,
                 event_fields: Optional[Dict[str, str]] = None):
        self._id = None
        self.name = name
        self.description = description
        if chunk_duration_us is not None:
            if chunk_duration != "":
                raise errors.ApiError("specify either chunk_duration or chunk_duration_us, not both")
            self.chunk_duration_us = chunk_duration_us
        else:
            self.chunk_duration_us = self._parse_time_duration(chunk_duration)
        if event_fields is None:
            self.event_fields = {}
        else:
            self.event_fields = validate_event_fields(event_fields)
        self.keep_us = keep_us

    @property
    def id(self) -> int:
        if self._id is None:
            raise errors.ApiError("this is a local model with no ID. See API docs")
        return self._id

    @id.setter
    def id(self, value: int):
        self._id = value

    def to_json(self) -> Dict:
        return {
            "id": self._id,
            "name": self.name,
            "description": self.description,
            "event_fields": self.event_fields,
            "chunk_duration_us": self.chunk_duration_us,
            "keep_us": self.keep_us
        }

    def __repr__(self):
        return "<joule.api.EventStream id=%r name=%r description=%r>" % (
            self._id, self.name, self.description
        )
        

    def _parse_time_duration(self, duration_str):
        # if duration is already a number interpret it as us
        try:
            return int(duration_str)
        except ValueError:
            pass
        # if it is empty use 0 (no hypertables)
        if duration_str == "":
            return 0
        #print(f"duration string: {duration_str}")
        # otherwise expect a time unit and compute accordingly
        match = re.fullmatch(r'^(\d+)([hdwmy])$', duration_str)
        if match is None:
            raise errors.ApiError("invalid duration "
                                  "use format #unit (eg 1w)")

        units = {
            'h': 60 * 60 * 1e6,  # hours
            'd': 24 * 60 * 60 * 1e6,  # days
            'w': 7 * 24 * 60 * 60 * 1e6,  # weeks
            'm': 4 * 7 * 24 * 60 * 60 * 1e6,  # months
            'y': 365 * 24 * 60 * 60 * 1e6  # years
        }
        unit = match.group(2)
        time = int(match.group(1))
        if time <= 0:
            raise errors.ApiError("invalid duration "
                                  "use format #unit (eg 1w), none or all")
        return int(time * units[unit])


def from_json(json: dict) -> EventStream:
    my_stream = EventStream()
    my_stream.id = json['id']
    my_stream.name = json['name']
    my_stream.description = json['description']
    my_stream.event_fields = json['event_fields']
    my_stream.keep_us = json['keep_us']
    # use a default of 0 for backwards compatability
    my_stream.chunk_duration_us = json.get("chunk_duration_us", 0)
    return my_stream


class EventStreamInfo:
    """
        API EventStreamInfo model. Received from :meth:`Node.event_stream_info` and should not be created directly.


        Parameters:
            start (int): timestamp in UNIX microseconds of the beginning of the first event
            end (int): timestamp in UNIX microsseconds of the end of the last data event
            event_count (int): number of events in the stream
            bytes (int): approximate size of the data on disk
            total_time (int): event stream duration in microseconds (start-end)

        """

    def __init__(self, start: Optional[int], end: Optional[int], event_count: int,
                 total_time: int = 0, bytes: int = 0):
        self.start = start
        self.end = end
        self.event_count = event_count
        self.bytes = bytes
        self.total_time = total_time

    def __repr__(self):
        return "<joule.api.EventStreamInfo start=%r end=%r event_count=%r, total_time=%r>" % (
            self.start, self.end, self.event_count, self.total_time)


def info_from_json(json) -> EventStreamInfo:
    if json is not None:

        return EventStreamInfo(json['start'],
                               json['end'],
                               json['event_count'],
                               json['total_time'],
                               json['bytes'])
    else:
        return EventStreamInfo(None,
                               None,
                               0,
                               0,
                               0)


async def event_stream_delete(session: BaseSession,
                              stream: EventStream | str | int) -> None:
    data = {}
    if type(stream) is EventStream:
        data["id"] = stream.id
    elif type(stream) is int:
        data["id"] = stream
    elif type(stream) is str:
        data["path"] = stream
    else:
        raise errors.InvalidEventStreamParameter()

    await session.delete(EndPoints.event, data)


async def event_stream_create(session: BaseSession,
                              stream: EventStream, folder: Folder | str | int) -> EventStream:
    data = {"stream": stream.to_json()}

    if type(folder) is Folder:
        data["dest_id"] = folder.id  # raises error if id is None
    elif type(folder) is int:
        data["dest_id"] = folder
    elif type(folder) is str:
        data["dest_path"] = folder
    else:
        raise errors.ApiError("Invalid folder datatype. Must be Folder, Path, or ID")

    resp = await session.post(EndPoints.event, json=data)
    return from_json(resp)


async def event_stream_info(session: BaseSession,
                            stream: EventStream | str | int) -> EventStreamInfo:
    data = {}

    if type(stream) is EventStream:
        data["id"] = stream.id  # raises error if id is None
    elif type(stream) is int:
        data["id"] = stream
    elif type(stream) is str:
        data["path"] = stream
    else:
        raise errors.InvalidEventStreamParameter()

    resp = await session.get(EndPoints.event, data)
    return info_from_json(resp['data_info'])


async def event_stream_get(session: BaseSession,
                           stream: EventStream | str | int,
                           create: bool = False,
                           description: str = "",
                           chunk_duration: str = "",
                           chunk_duration_us: Optional[int] = None,
                           event_fields=None) -> EventStream:
    data = {}

    if type(stream) is EventStream:
        data["id"] = stream.id
    elif type(stream) is int:
        data["id"] = stream
    elif type(stream) is str:
        data["path"] = stream
    else:
        raise errors.InvalidEventStreamParameter()

    try:
        resp = await session.get(EndPoints.event, data)
    except errors.ApiError as e:
        # pass the error if the stream should not or cannot be created
        if not create or type(stream) is not str:
            raise e
        name = stream.split('/')[-1]
        path = '/'.join(stream.split('/')[:-1])
        event_stream = EventStream(name, description,
                                   chunk_duration=chunk_duration,
                                   chunk_duration_us=chunk_duration_us,
                                   event_fields=event_fields)
        return await event_stream_create(session, event_stream, path)
    return from_json(resp)


async def event_stream_update(session: BaseSession,
                              stream: EventStream) -> None:
    # validate the event fields before sending
    validate_event_fields(stream.event_fields)
    await session.put(EndPoints.event, {"id": stream.id,
                                      "stream": stream.to_json()})


async def event_stream_move(session: BaseSession,
                            source: EventStream | str | int,
                            destination: Folder | str | int) -> None:
    data = {}

    if type(source) is EventStream:
        data["src_id"] = source.id
    elif type(source) is int:
        data["src_id"] = source
    elif type(source) is str:
        data["src_path"] = source
    else:
        raise errors.ApiError("Invalid source datatype. Must be EventStream, Path, or ID")

    if type(destination) is Folder:
        data["dest_id"] = destination.id
    elif type(destination) is int:
        data["dest_id"] = destination
    elif type(destination) is str:
        data["dest_path"] = destination
    else:
        raise errors.ApiError("Invalid destination datatype. Must be Folder, Path, or ID")
    await session.put(EndPoints.event_move, data)


async def event_stream_write(session: BaseSession,
                             stream: EventStream | str | int,
                             events: List[Event]) -> List[Event]:
    data = {}
    if type(stream) is EventStream:
        data["id"] = stream.id
    elif type(stream) is int:
        data["id"] = stream
    elif type(stream) is str:
        data["path"] = stream
    else:
        raise errors.InvalidEventStreamParameter()
    # get the node UUID to make sure events can be associated with the correct node during copy operations
    resp = await session.get(EndPoints.version_json)
    node_uuid = resp['uuid']
    # post events in blocks
    rx_events = []
    for idx in range(0, len(events), 500):
        chunk = events[idx:idx + 500]
        data['events'] = [e.to_json(destination_node_uuid=node_uuid) for e in events[idx:idx + 500]]
        resp = await session.post(EndPoints.event_data, data)
        rx_events += [event_from_json(e, node_uuid=node_uuid) for e in resp["events"]]
        # copy the ids over
        for i in range(len(chunk)):
            chunk[i].id = rx_events[i].id
    return events


async def event_stream_count(session: BaseSession,
                             stream: EventStream | str | int,
                             start_time: Optional[int],
                             end_time: Optional[int],
                             json_filter,
                             include_on_going_events) -> List[Event]:
    params = {}
    if type(stream) is EventStream:
        params["id"] = stream.id
    elif type(stream) is int:
        params["id"] = stream
    elif type(stream) is str:
        params["path"] = stream
    else:
        raise errors.InvalidEventStreamParameter()
    if start_time is not None:
        params['start'] = int(start_time)
    if end_time is not None:
        params['end'] = int(end_time)
    if json_filter is not None:
        params['filter'] = json_filter
    if include_on_going_events:
        params['include-ongoing-events'] = 1
    resp = await session.get(EndPoints.event_data_count, params)
    return resp["count"]

async def event_stream_read_list(session: BaseSession,
                                stream: EventStream | str | int,
                                start_time: Optional[int],
                                end_time: Optional[int],
                                json_filter,
                                include_on_going_events,
                                limit)-> List[Event]:
    if limit <= 0:
        raise errors.ApiError("limit must be an integer > 0")
    all_events = []
    async for events in event_stream_read(session=session, stream=stream, start_time=start_time, end_time=end_time, json_filter=json_filter,
                                         include_on_going_events=include_on_going_events, block_size=EVENT_READ_BLOCK_SIZE):
        all_events += events
        if len(all_events) >= limit:
            log.warning(f"Read limit of {limit} events reached before end of stream")
            break
    return all_events[:limit]
        
async def event_stream_read(session: BaseSession,
                            stream: EventStream | str | int,
                            start_time: Optional[int],
                            end_time: Optional[int],
                            json_filter,
                            include_on_going_events,
                            block_size: Optional[int]):
    params = {}
    if type(stream) is EventStream:
        params["id"] = stream.id
    elif type(stream) is int:
        params["id"] = stream
    elif type(stream) is str:
        params["path"] = stream
    else:
        raise errors.InvalidEventStreamParameter()
    if start_time is not None:
        params['start'] = int(start_time)
    if end_time is not None:
        params['end'] = int(end_time)
    if block_size is None:
        params['limit'] = EVENT_READ_BLOCK_SIZE
    else:
        params['limit'] = block_size
    params['return-subset'] = 1
    if json_filter is not None:
        params['filter'] = json_filter
    if include_on_going_events:
        params['include-ongoing-events'] = 1

    # get the node UUID to make sure events can be associated with the correct node during copy operations
    resp = await session.get(EndPoints.version_json)
    node_uuid = resp['uuid']
    
    # read events in blocks and provide to client as a generator
    while True:
        resp = await session.get(EndPoints.event_data, params)
        events_json = resp["events"]
        events = [event_from_json(e, node_uuid=node_uuid) for e in events_json]
        # if no events are returned, we are done
        if len(events)==0:
            break
        #print(f"Retrieved {len(events)} from {ts2h(events[0].start_time)} to {ts2h(events[-1].start_time)}")
        if block_size is not None:
            yield events
        else:
            for event in events:
                yield event
        # update the start time to get the next block
        params['start'] = events[-1].start_time+1
        if end_time is not None and params['start'] >= end_time:
            break
        params['include-ongoing-events'] = 0 # should always be set false after the first pass

async def event_stream_remove(session: BaseSession,
                              stream: EventStream | str | int,
                              start_time: Optional[int] = None,
                              end_time: Optional[int] = None,
                              json_filter=None) -> None:
    params = {}
    if type(stream) is EventStream:
        params["id"] = stream.id
    elif type(stream) is int:
        params["id"] = stream
    elif type(stream) is str:
        params["path"] = stream
    else:
        raise errors.InvalidEventStreamParameter()
    if start_time is not None:
        params['start'] = int(start_time)
    if end_time is not None:
        params['end'] = int(end_time)
    if json_filter is not None:
        params['filter'] = json_filter
    await session.delete(EndPoints.event_data, params)



import datetime

from tests.helpers import DbTestCase
from joule.models import (Base, DataStream, Folder,
                          Element, Follower)
from joule.errors import ConfigurationError
from joule.services import parse_pipe_config


class TestParsePipeConfig(DbTestCase):

    def setUp(self):
        super().setUp()
        # /test/stream1:float32[e0,e1,e2]
        folder_test = Folder(name="test",
                             updated_at=datetime.datetime.now())
        stream1 = DataStream(name="stream1", keep_us=100,
                             datatype=DataStream.DATATYPE.FLOAT32,
                             updated_at=datetime.datetime.now())
        stream1.elements = [Element(name="e%d" % x, index=x, default_min=1) for x in range(3)]
        folder_test.data_streams.append(stream1)

        # /test/deeper/stream2:int16[e0,e1]
        folder_deeper = Folder(name="deeper",
                               updated_at=datetime.datetime.now())
        stream2 = DataStream(name="stream2", datatype=DataStream.DATATYPE.INT16,
                             updated_at=datetime.datetime.now())
        stream2.elements = [Element(name="e%d" % x, index=x) for x in range(2)]
        folder_deeper.data_streams.append(stream2)
        folder_deeper.parent = folder_test

        root = Folder(name="root",
                      updated_at=datetime.datetime.now())
        root.children = [folder_test]

        remote_node = Follower(name="remote.node", key="api_key", location="https://remote.com:8088")
        self.db.add(root)
        self.db.add(remote_node)
        self.db.commit()

    def test_parses_new_stream_configs(self):
        my_stream = parse_pipe_config.run("/test/new_stream:int16[x,y]", self.db)
        # ensure stream entered the database
        self.assertEqual(my_stream, self.db.query(DataStream).filter_by(name="new_stream").one())
        # check stream datatype
        self.assertEqual(my_stream.datatype, DataStream.DATATYPE.INT16)
        # check elements
        self.assertEqual(len(my_stream.elements), 2)
        for elem in my_stream.elements:
            self.assertIn(elem.name, ['x', 'y'])
        # check parent folder
        self.assertEqual(my_stream.folder.name, "test")

    def test_parses_existing_stream_configs(self):
        existing_stream = self.db.query(DataStream).filter_by(name="stream2").one()
        my_stream = parse_pipe_config.run("/test/deeper/stream2:int16[e0,e1]", self.db)
        # retrieved the existsing stream from the database
        self.assertEqual(my_stream, existing_stream)

    def test_parses_links_to_existing_configs(self):
        existing_stream = self.db.query(DataStream).filter_by(name="stream2").one()
        my_stream = parse_pipe_config.run("/test/deeper/stream2", self.db)
        # retrieved the existsing stream from the database
        self.assertEqual(my_stream, existing_stream)

    def test_parses_remote_stream_configs(self):
        my_stream = parse_pipe_config.run("remote.node /path/to/stream:float32[x,y]", self.db)
        self.assertTrue(my_stream.is_remote)
        self.assertEqual(my_stream.remote_node, "remote.node")
        self.assertEqual(my_stream.remote_path, "/path/to/stream")

    def test_errors_on_invalid_configs(self):
        bad_configs = ["no/leading/slash", "/", "/invalid/datatype:uknown[x,y]",
                       "/bad/syntax:int[", "/bad/syntax:missing", "remote.node /missing/inline/config"]
        for config in bad_configs:
            with self.assertRaisesRegex(ConfigurationError, "invalid"):
                parse_pipe_config.run(config, self.db)

    def test_errors_on_stream_datatype_conflict(self):
        with self.assertRaisesRegex(ConfigurationError, "exists"):
            parse_pipe_config.run("/test/deeper/stream2:int64[e0,e1]", self.db)

    def test_errors_on_stream_element_name_conflict(self):
        with self.assertRaisesRegex(ConfigurationError, "exists"):
            parse_pipe_config.run("/test/deeper/stream2:int16[e0,e1,e2]", self.db)
        with self.assertRaisesRegex(ConfigurationError, "exists"):
            parse_pipe_config.run("/test/deeper/stream2:int16[e0,x]", self.db)

    def test_errors_if_unconfigured_stream_not_in_db(self):
        with self.assertRaisesRegex(ConfigurationError, "/test/deeper/new_stream"):
            parse_pipe_config.run("/test/deeper/new_stream", self.db)



from click.testing import CliRunner
import unittest
import asyncio
import warnings
from tests.cli.fake_joule import FakeJoule, FakeJouleTestCase
from joule.cli import main



class TestMasterAdd(FakeJouleTestCase):

    def test_add_user(self):
        server = FakeJoule()
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, "master add user johndoe".split(" "))
        output = result.output
        # make sure the username and key are printed
        self.assertIn("fakekey", output)
        self.assertIn("johndoe", output)
        # make sure the server got the right parameters
        params = self.msgs.get()
        self.assertEqual(params['identifier'], "johndoe")
        self.assertEqual(params['master_type'], "user")
        # make sure everything stops cleanly
        self.assertEqual(result.exit_code, 0)
        self.stop_server()

    def test_add_joule(self):
        server = FakeJoule()
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, "master add joule node1".split(" "))
        output = result.output
        # make sure the node name is printed
        self.assertIn("node1", output)
        # make sure the server got the right parameters
        params = self.msgs.get()
        self.assertEqual(params['identifier'], "node1")
        self.assertEqual(params['master_type'], "joule")
        # make sure everything stops cleanly
        self.assertEqual(result.exit_code, 0)
        self.stop_server()

    def test_add_lumen(self):
        server = FakeJoule()
        self.start_server(server)
        runner = CliRunner()
        # The first node joining a lumen requires a user
        result = runner.invoke(main, "master add lumen node1".split(" "),
                               input="John\nDoe\njohndoe@email.com\npassword\npassword")
        output = result.output
        # make sure the node name is printed
        self.assertIn("node1", output)
        # make sure the server got the right parameters
        params = self.msgs.get()
        self.assertEqual(params["first_name"], "John")
        self.assertEqual(params["last_name"], "Doe")
        self.assertEqual(params["email"], "johndoe@email.com")
        self.assertEqual(params["password"], "password")

        #loop = asyncio.new_event_loop()
        #asyncio.set_event_loop(loop)
        # Subsequent nodes need an auth key
        result = runner.invoke(main, "master add lumen node2".split(" "),
                               input="AC3412\n")
        # make sure the server got the right parameters
        params = self.msgs.get()
        self.assertEqual(params["auth_key"], "AC3412")
        output = result.output
        self.assertIn("node2", output)
        # make sure everything stops cleanly
        self.assertEqual(result.exit_code, 0)
        self.stop_server()

    def test_bad_arguments(self):
        runner = CliRunner()
        result = runner.invoke(main, ['master', 'add','invalid','node2'])
        self.assertIn('Error', result.output)
        self.assertEqual(result.exit_code, 2)

    def test_when_server_returns_invalid_data(self):
        server = FakeJoule()
        server.response = "notjson"
        server.stub_master = True
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, "master add user node2".split(" "))
        self.assertIn('Error', result.output)
        self.assertEqual(result.exit_code, 1)
        self.stop_server()

    def test_when_server_returns_error_code(self):
        server = FakeJoule()
        error_msg = "test error"
        error_code = 500
        server.response = error_msg
        server.http_code = error_code
        server.stub_master = True
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, "master add user node2".split(" "))
        self.assertIn(error_msg, result.output)
        self.assertEqual(result.exit_code, 1)
        self.stop_server()



from click.testing import CliRunner
import os
import warnings
import json
from tests.cli.fake_joule import FakeJoule, FakeJouleTestCase
from joule.models import Proxy
from joule.cli import main
from ..fake_joule import print_result_on_error


class TestFollowerList(FakeJouleTestCase):

    def test_lists_proxies(self):
        server = FakeJoule()
        proxies = [Proxy(name="P1", uuid=4,url="http://127.0.0.1:8080"),
                     Proxy(name="P2", uuid=5,url="http://127.0.0.1:8081")]
        server.response = json.dumps([f.to_json() for f in proxies])
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['proxy', 'list'])
        print_result_on_error(result)
        self.assertEqual(result.exit_code, 0)
        output = result.output
        # make sure followers are listed
        for name in ['P1', 'P2']:
            self.assertIn(name, output)
        self.stop_server()

    def test_when_server_returns_error_code(self):
        server = FakeJoule()
        error_msg = "test error"
        error_code = 500
        server.response = error_msg
        server.http_code = error_code
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['proxy', 'list'])
        self.assertIn('%d' % error_code, result.output)
        self.assertIn(error_msg, result.output)
        self.assertEqual(result.exit_code, 1)
        self.stop_server()



from tests.api import mock_session
import random
import unittest

from joule.api.node import TcpNode
from joule import errors

from joule.api.data_stream import DataStream
from joule.api.folder import Folder
from .helpers import build_stream
from joule.constants import EndPoints

class TestFolderApi(unittest.IsolatedAsyncioTestCase):

    async def asyncSetUp(self) -> None:
        # no URL or event loop
        self.node = TcpNode('mock_node', 'http://url', 'api_key')
        self.session = mock_session.MockSession()
        self.node.session = self.session

    async def test_folder_move(self):
        # can move by ID
        await self.node.folder_move(1, 2)
        self.assertEqual(self.session.method, 'PUT')
        self.assertEqual(self.session.path, EndPoints.folder_move)
        self.assertEqual(self.session.request_data,
                         {'src_id': 1, 'dest_id': 2})
        # can move by path
        await self.node.folder_move('/a/path', '/b/path')
        self.assertEqual(self.session.method, 'PUT')
        self.assertEqual(self.session.path, EndPoints.folder_move)
        self.assertEqual(self.session.request_data,
                         {'src_path': '/a/path', 'dest_path': '/b/path'})

        # can move by Folder
        src = Folder()
        src.id = 1
        dest = Folder()
        dest.id = 2
        await self.node.folder_move(src, dest)
        self.assertEqual(self.session.path, EndPoints.folder_move)
        self.assertEqual(self.session.request_data,
                         {'src_id': 1, 'dest_id': 2})

    async def test_folder_move_errors(self):
        self.session.method = None
        with self.assertRaises(errors.ApiError):
            await self.node.folder_move(DataStream(), '/b/path')
        self.assertIsNone(self.session.method)

        with self.assertRaises(errors.ApiError):
            await self.node.folder_move('/a/path', DataStream())
        self.assertIsNone(self.session.method)

    async def test_folder_delete(self):
        # can delete by ID
        await self.node.folder_delete(1, recursive=False)
        self.assertEqual(self.session.method, 'DELETE')
        self.assertEqual(self.session.path, EndPoints.folder)
        self.assertEqual(self.session.request_data, {'id': 1, 'recursive': 0})
        # can delete by path
        await self.node.folder_delete('/a/path', recursive=True)
        self.assertEqual(self.session.method, 'DELETE')
        self.assertEqual(self.session.path, EndPoints.folder)
        self.assertEqual(self.session.request_data, {'path': '/a/path', 'recursive': 1})

        # can delete by Folder
        src = Folder()
        src.id = 1
        await self.node.folder_delete(src, recursive=False)
        self.assertEqual(self.session.method, 'DELETE')
        self.assertEqual(self.session.path, EndPoints.folder)
        self.assertEqual(self.session.request_data, {'id': 1, 'recursive': 0})

    async def test_folder_delete_errors(self):
        self.session.method = None
        with self.assertRaises(errors.ApiError):
            await self.node.folder_delete(DataStream(), recursive=False)
        self.assertIsNone(self.session.method)

    async def test_folder_get(self):
        target = build_folder(3)
        self.session.response_data = target.to_json()
        folder = await self.node.folder_get(1)
        self.assertEqual(folder.to_json(), target.to_json())
        self.assertEqual(self.session.method, 'GET')
        self.assertEqual(self.session.path, EndPoints.folder)
        self.assertEqual(self.session.request_data, {'id': 1})
        # can get by path
        await self.node.folder_get('/a/path')
        self.assertEqual(self.session.method, 'GET')
        self.assertEqual(self.session.path, EndPoints.folder)
        self.assertEqual(self.session.request_data, {'path': '/a/path'})

        # can get by Folder
        src = Folder()
        src.id = 1
        await self.node.folder_get(src)
        self.assertEqual(self.session.path, EndPoints.folder)
        self.assertEqual(self.session.request_data, {'id': 1})

    async def test_folder_get_errors(self):
        self.session.method = None
        with self.assertRaises(errors.ApiError):
            await self.node.folder_get(DataStream())
        self.assertIsNone(self.session.method)

    async def test_local_folder_has_no_id(self):
        folder = Folder()
        with self.assertRaises(errors.ApiError):
            folder.id


def build_folder(depth) -> Folder:
    # can get by ID
    target = Folder()
    target.id = idgen()
    target.name = 'f'
    target.description = ''
    for n in range(random.randint(2, 4)):
        s = build_stream('s%d' % n, 'float32[x,y,z]')
        s.id = idgen()
        target.data_streams.append(s)
    if depth > 0:
        for n in range(random.randint(2, 3)):
            target.children.append(build_folder(depth - 1))
    return target


id_seq = 0


def idgen():
    global id_seq
    id_seq += 1
    return id_seq



#!/usr/bin/env python3
import time
import socket
import subprocess
import os
import asyncio
import aiohttp
from joule import api
from joule.api.data_stream import DataStream, Element
import unittest

test_case = unittest.TestCase()

async def main():
    time.sleep(4)  # wait for jouled to boot and modules to start
    # start a standalone visualizer module with mock data
    module_config = os.path.join(os.path.dirname(__file__), "standalone_app.conf")
    app_port = _find_port()
    standalone_app= subprocess.Popen(
        f"coverage run --rcfile=/joule/.coveragerc -m joule.client.builtins.visualizer --module_config={module_config} --live --port={app_port}".split(" "))
    time.sleep(4)  # let the module start
    await validate_operation(standalone_app, app_port)
    standalone_app.kill()

async def validate_operation(standalone_app, app_port):
    # get the node
    node = api.get_node()
    integrated_app = await node.module_get("Data App")

    async with aiohttp.ClientSession() as session:
        # the title from the module configuration file should be on the page
        async with session.get(f"http://localhost:{app_port}") as response:
            html = await response.text()
            test_case.assertIn("Standalone App", html)

        # cannot get the integrated module without permission
        async with session.get(f"http://localhost/joule/app/m{integrated_app.id}/") as response:
            test_case.assertEqual(response.status, 403)
            
        # can get the integrated module with permission, title should be on the page
        async with session.get(f"http://localhost/joule/app/m{integrated_app.id}/", headers={"X-API-KEY": node._key}) as response:
            test_case.assertEqual(response.status, 200)
            html = await response.text()
            test_case.assertIn("Quick Start Data Pipeline", html)

        # resetting the module should clear the min/max ranges
        async with session.get(f"http://localhost:{app_port}/data.json") as response:
            test_case.assertEqual(response.status, 200)
            data = await response.json()
            for element in data:
                test_case.assertLessEqual(element["min"], element["value"])
                test_case.assertGreaterEqual(element["max"], element["value"])

        async with session.post(f"http://localhost:{app_port}/reset.json") as response:
            test_case.assertEqual(response.status, 200)

        async with session.get(f"http://localhost:{app_port}/data.json") as response:
            test_case.assertEqual(response.status, 200)
            data = await response.json()
            for element in data:
                test_case.assertEqual(element["min"], "&mdash;")
                test_case.assertEqual(element["max"], "&mdash;")
           

    await node.close()
    
def _find_port():
    # return an available TCP port, check to make sure it is available
    # before returning
    for port in range(8000, 9000):
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
            try:
                sock.bind(('localhost', port))
                return port
            except OSError:
                continue


if __name__ == "__main__":
    asyncio.run(main())
    print("OK")


from tests import helpers
from unittest import mock
import numpy as np
import asyncio
from icecream import ic
from joule.models.pipes import LocalPipe, OutputPipe
from joule.errors import EmptyPipeError, PipeError


class TestLocalPipe(helpers.AsyncTestCase):

    def test_pipes_numpy_arrays(self):
        """writes to pipe sends data to reader and any subscribers"""
        LAYOUT = "int16_2"
        LENGTH = 1003
        ''
        my_pipe = LocalPipe(LAYOUT, name="pipe")
        # test timeouts, since the writer is slow
        my_pipe.TIMEOUT_INTERVAL = 0.05
        subscriber_pipe = LocalPipe(LAYOUT)
        subscriber_pipe2 = LocalPipe(LAYOUT)
        my_pipe.subscribe(subscriber_pipe)
        my_pipe.subscribe(subscriber_pipe2)
        test_data = helpers.create_data(LAYOUT, length=LENGTH)
        reader_rx_data = np.empty(test_data.shape, test_data.dtype)
        subscriber_rx_data = np.empty(test_data.shape, test_data.dtype)

        #    print(test_data['data'][:,1])

        async def writer():
            for block in helpers.to_chunks(test_data, 270):
                await asyncio.sleep(0.1)
                await my_pipe.write(block)
            await my_pipe.close()

        async def reader():
            # note this is kind of delicate, you can only read data twice
            # if the pipe is closed so it might look like you didn't finish all
            # the data if the last read ignores data past blk_size
            blk_size = 600
            rx_idx = 0
            while await my_pipe.not_empty():
                # consume data in pipe
                data_chunk = await my_pipe.read()
                rows_used = min(len(data_chunk), blk_size)
                reader_rx_data[rx_idx:rx_idx + rows_used] = data_chunk[:rows_used]
                rx_idx += rows_used
                my_pipe.consume(rows_used)

        async def subscriber():
            rx_idx = 0
            while await subscriber_pipe.not_empty():
                # consume data in pipe
                data_chunk = await subscriber_pipe.read()
                subscriber_rx_data[rx_idx:rx_idx + len(data_chunk)] = data_chunk
                rx_idx += len(data_chunk)
                subscriber_pipe.consume(len(data_chunk))

        async def runner():
            await asyncio.gather(writer(),
                                 reader(),
                                 subscriber())
        asyncio.run(runner())

        data = subscriber_pipe2.read_nowait()
        np.testing.assert_array_equal(test_data, data)
        np.testing.assert_array_equal(test_data, reader_rx_data)
        np.testing.assert_array_equal(test_data, subscriber_rx_data)

    def test_read_data_must_be_consumed(self):
        # writes to pipe sends data to reader and any subscribers
        LAYOUT = "float32_2"
        LENGTH = 500
        UNCONSUMED_ROWS = 4
        my_pipe = LocalPipe(LAYOUT)
        chunk1 = helpers.create_data(LAYOUT, length=LENGTH)
        chunk2 = helpers.create_data(LAYOUT, length=LENGTH)
        chunk3 = helpers.create_data(LAYOUT, length=LENGTH)

        my_pipe.write_nowait(chunk1)

        async def reader():
            await my_pipe.read()
            my_pipe.consume(0)
            # should get the same data back on the next read
            # add a second copy of the test data
            await my_pipe.write(chunk2)
            rx_data = await my_pipe.read()
            # two copies of the data now
            np.testing.assert_array_equal(chunk1, rx_data[:len(chunk1)])
            np.testing.assert_array_equal(chunk2, rx_data[len(chunk1):])
            # write another copy but consume the first
            my_pipe.consume(len(chunk1))
            await my_pipe.write(chunk3)
            rx_data = await my_pipe.read()
            # two copies of the data now
            np.testing.assert_array_equal(chunk2, rx_data[:len(chunk2)])
            np.testing.assert_array_equal(chunk3, rx_data[len(chunk2):])
            my_pipe.consume(len(chunk2))
            await my_pipe.close()
            # now a read should return immediately with the unconsumed data
            rx_data = await my_pipe.read()
            np.testing.assert_array_equal(chunk3, rx_data)
            # the pipe should be empty but still return the old data
            rx_data = await my_pipe.read()
            np.testing.assert_array_equal(chunk3, rx_data)
            # only after consuming the remaining data does it raise an exception
            my_pipe.consume(len(rx_data))
            # another read should cause an exception (even if the data hasn't been consumed)
            with self.assertRaises(EmptyPipeError):
                await my_pipe.read()
            # the pipe should be empty
            self.assertTrue(await my_pipe.is_empty())

        asyncio.run(reader())

    def test_consumes_data(self):
        LAYOUT = "int16_2"
        LENGTH = 500
        test_data = helpers.create_data(LAYOUT, length=LENGTH)
        my_pipe = LocalPipe("int16_2")
        my_pipe.write_nowait(test_data)
        my_pipe.close_nowait()
        async def reader():
            data = await my_pipe.read()
            assert len(data) == LENGTH
            my_pipe.consume(100)
            data = await my_pipe.read()
            assert len(data) == LENGTH - 100
            my_pipe.consume_all()
            assert await my_pipe.is_empty()

        asyncio.run(reader())

    def test_nowait_read_writes(self):
        LAYOUT = "int16_2"
        LENGTH = 500
        ''
        my_pipe = LocalPipe(LAYOUT)
        my_subscriber = LocalPipe(LAYOUT)
        my_pipe.subscribe(my_subscriber)
        test_data = helpers.create_data(LAYOUT, length=LENGTH)
        my_pipe.write_nowait(test_data)
        result = my_pipe.read_nowait()
        # read data should match written data
        np.testing.assert_array_almost_equal(test_data['data'], result['data'])
        np.testing.assert_array_almost_equal(
            test_data['timestamp'], result['timestamp'])
        # subscriber should have a copy too
        result = my_subscriber.read_nowait()
        np.testing.assert_array_almost_equal(test_data['data'], result['data'])
        np.testing.assert_array_almost_equal(
            test_data['timestamp'], result['timestamp'])

    def test_invalid_write_nowait_inputs(self):
        LAYOUT = "int16_2"
        ''
        my_pipe = LocalPipe(LAYOUT, name="testpipe")
        with self.assertLogs(level="INFO"):
            my_pipe.write_nowait(np.array([[]]))
        with self.assertRaises(PipeError):
            my_pipe.write_nowait([1, 2, 3])

    def test_different_format_writes(self):
        LAYOUT = "int16_2"
        ''
        my_pipe = LocalPipe(LAYOUT, name="testpipe")

        test_data = helpers.create_data(LAYOUT, length=4, step=1, start=106)

        async def write():
            # write unstructured numpy arrays
            await my_pipe.write(np.array([[1, 1, 1]]))
            await my_pipe.write(np.array([[2, 2, 2],
                                          [3, 3, 3]]))

            # logs empty writes
            with self.assertLogs(level="INFO") as logs:
                await my_pipe.write(np.array([[]]))

            # errors on invalid write types
            bad_data = [[100, 1, 2], 'invalid', 4, None, np.array([4, 8])]
            for data in bad_data:
                with self.assertRaises(PipeError):
                    await my_pipe.write(data)

            # write structured numpy arrays
            await my_pipe.write(test_data)

        ''
        asyncio.run(write())
        result = my_pipe.read_nowait(flatten=True)
        np.testing.assert_array_equal(result[:3], [[1, 1, 1],
                                                   [2, 2, 2],
                                                   [3, 3, 3]])
        my_pipe.consume(3)
        result = my_pipe.read_nowait()
        np.testing.assert_array_equal(result, test_data)

    def test_caching(self):
        LAYOUT = "float32_2"
        LENGTH = 128
        CACHE_SIZE = 40
        my_pipe = LocalPipe(LAYOUT)
        my_pipe.enable_cache(CACHE_SIZE)
        test_data = helpers.create_data(LAYOUT, length=LENGTH, start=1, step=1)

        async def writer():
            for block in helpers.to_chunks(test_data, 4):
                await asyncio.sleep(.1)
                await my_pipe.write(block)
            # closing the interval should flush the data
            await my_pipe.close_interval()
            # add a dummy section after the interval break
            await my_pipe.write(np.ones((35, 3)))
            await my_pipe.flush_cache()

        num_reads = 0

        async def reader():
            nonlocal num_reads
            index = 0
            while True:
                data = await my_pipe.read()
                my_pipe.consume(len(data))
                # make sure the data is correct
                np.testing.assert_array_equal(test_data[index:index + len(data)],
                                              data)
                num_reads += 1
                index += len(data)

                if index == len(test_data):
                    break
            # now get the dummy section after the interval break
            data = await my_pipe.read(flatten=True)
            np.testing.assert_array_equal(data, np.ones((35, 3)))

        async def runner():
            await asyncio.gather(reader(), writer())
        asyncio.run(runner())
        self.assertLessEqual(num_reads, np.ceil(LENGTH / CACHE_SIZE))

    def test_nowait_read_empties_queue(self):
        LAYOUT = "int16_2"
        LENGTH = 50
        my_pipe = LocalPipe(LAYOUT)
        test_data = helpers.create_data(LAYOUT, length=LENGTH)

        async def writer():
            for row in test_data:
                await my_pipe.write(np.array([row]))

        ''
        asyncio.run(writer())

        result = my_pipe.read_nowait()

        np.testing.assert_array_almost_equal(test_data['data'], result['data'])
        np.testing.assert_array_almost_equal(
            test_data['timestamp'], result['timestamp'])

    def test_nowait_read_does_not_block(self):
        LAYOUT = "int16_2"
        LENGTH = 50
        my_pipe = LocalPipe(LAYOUT)
        test_data = helpers.create_data(LAYOUT, length=LENGTH)
        my_pipe.write_nowait(test_data)
        self.assertEqual(len(my_pipe.read_nowait()), LENGTH)
        # another read executes immediately because there is unconsumed data
        self.assertEqual(len(my_pipe.read_nowait()), LENGTH)
        # now consume the data and the next call to read execute immediately
        # even though there is no data to be returned
        my_pipe.consume(LENGTH)
        self.assertEqual(0, len(my_pipe.read_nowait(flatten=True)))
        return

    def test_handles_flat_and_structured_arrays(self):
        # converts flat arrays to structured arrays and returns
        #   either flat or structured arrays depending on [flatten] parameter
        LAYOUT = "float64_1"
        LENGTH = 1000
        my_pipe = LocalPipe(LAYOUT)
        test_data = helpers.create_data(LAYOUT, length=LENGTH)
        flat_data = np.c_[test_data['timestamp'][:, None], test_data['data']]

        async def reader():
            sdata = await my_pipe.read()
            fdata = await my_pipe.read(flatten=True)
            np.testing.assert_array_almost_equal(
                sdata['timestamp'], test_data['timestamp'])
            np.testing.assert_array_almost_equal(
                sdata['data'], test_data['data'])
            np.testing.assert_array_almost_equal(fdata, flat_data)
            np.testing.assert_array_almost_equal(fdata, flat_data)

        asyncio.run(my_pipe.write(flat_data))
        asyncio.run(my_pipe.close())
        asyncio.run(reader())

    def test_raises_consume_errors(self):
        LAYOUT = "int32_3"
        LENGTH = 1000
        my_pipe = LocalPipe(LAYOUT)
        test_data = helpers.create_data(LAYOUT, length=LENGTH)
        my_pipe.write_nowait(test_data)
        read_data = my_pipe.read_nowait()
        # can't consume more than was read
        with self.assertRaises(PipeError) as e:
            my_pipe.consume(len(read_data) + 1)
        self.assertTrue('consume' in str(e.exception))
        # can't consume less than zero
        with self.assertRaises(PipeError) as e:
            my_pipe.consume(-1)
        self.assertTrue('negative' in str(e.exception))

    def test_handles_interval_breaks(self):
        LAYOUT = "int32_3"
        LENGTH = 1000
        my_pipe = LocalPipe(LAYOUT, name="pipe")
        test_data1 = helpers.create_data(LAYOUT, length=LENGTH)
        test_data2 = helpers.create_data(LAYOUT, length=LENGTH)
        my_pipe.write_nowait(test_data1)
        my_pipe.close_interval_nowait()
        my_pipe.write_nowait(test_data2)

        async def reader():
            # read the first interval
            read_data = await my_pipe.read()
            self.assertTrue(my_pipe.end_of_interval)
            my_pipe.consume(len(read_data) - 20)
            np.testing.assert_array_equal(test_data1, read_data)

            # read the second interval
            read_data = await my_pipe.read()
            self.assertFalse(my_pipe.end_of_interval)
            self.assertEqual(len(read_data), len(test_data2) + 20)
            my_pipe.consume(len(read_data))
            np.testing.assert_array_equal(test_data2, read_data[20:])

        asyncio.run(reader())

    def test_nowait_function_exceptions(self):
        pipe = LocalPipe(layout="float32_3")
        subscriber = LocalPipe(layout="float32_3", close_cb=mock.Mock())
        fd_subscriber = OutputPipe()
        # cannot close_nowait if there is a callback
        with self.assertRaises(PipeError):
            subscriber.close_nowait()

        # can write_nowait if only localpipe subscribers
        pipe.subscribe(subscriber)
        pipe.write_nowait(np.ones((100, 4)))
        # cannot if any subscribers are not localpipes
        pipe.subscribe(fd_subscriber)
        with self.assertRaises(PipeError):
            pipe.write_nowait(np.ones((100, 4)))

        # cannot close_nowait if there are subscribers
        pipe.subscribe(subscriber)
        with self.assertRaises(PipeError):
            pipe.close_nowait()



import click

from joule.cli.lazy_group import LazyGroup

@click.group(name="admin",
             cls=LazyGroup,
             lazy_subcommands={"initialize": "joule.cli.admin.initialize.admin_initialize",
                               "erase": "joule.cli.admin.erase.admin_erase",
                               "authorize": "joule.cli.admin.authorize.admin_authorize"})
def admin():
    """Administer the local node."""
    pass  


from sqlalchemy import create_engine
from sqlalchemy.orm import Session
import unittest
import tempfile
import logging
import os
import warnings
import datetime

from tests.helpers import DbTestCase
from joule.models import (Base, DataStream, Folder, Element)
from joule.services import load_streams

logger = logging.getLogger('joule')

warnings.simplefilter('always')


class TestConfigureStreams(DbTestCase):

    def test_merges_config_and_db_streams(self):
        """e2e stream configuration service test"""

        # /test/stream1:float32_3
        folder_test = Folder(name="test",
                             updated_at=datetime.datetime.now())
        stream1 = DataStream(name="stream1", keep_us=100,
                             datatype=DataStream.DATATYPE.FLOAT32,
                             updated_at=datetime.datetime.now())
        stream1.elements = [Element(name="e%d" % x, index=x, default_min=1) for x in range(3)]
        folder_test.data_streams.append(stream1)

        # /test/deeper/stream2: int16_2
        folder_deeper = Folder(name="deeper",
                               updated_at=datetime.datetime.now())
        stream2 = DataStream(name="stream2", datatype=DataStream.DATATYPE.INT16,
                             updated_at=datetime.datetime.now())
        stream2.elements = [Element(name="e%d" % x, index=x) for x in range(2)]
        folder_deeper.data_streams.append(stream2)
        folder_deeper.parent = folder_test

        # /test/deeper/stream3: int16_2
        stream3 = DataStream(name="stream3", datatype=DataStream.DATATYPE.INT16,
                             updated_at=datetime.datetime.now())
        stream3.elements = [Element(name="e%d" % x, index=x) for x in range(2)]
        folder_deeper.data_streams.append(stream3)

        root = Folder(name="root",
                      updated_at=datetime.datetime.now())
        root.children = [folder_test]
        self.db.add(root)

        self.db.commit()
        configs = [
            # /test/stream1:float32_3 <different element configs and keep>
            """
            [Main] 
              name = stream1
              path = /test
              datatype = float32
              keep = all
            [Element1]
              name=new_e1
              display_type=discrete
            [Element2]
              name=new_e2
              display_type=event
            [Element3]
              name=new_e3
              default_min=-10
            """,
            # /new/path/stream4: int16_2 <a new stream>
            """
            [Main]
              name = stream4
              path = /new/path
              datatype = int16
            [Element1]
              name=1
            [Element2]
              name=2
            """,
            # /new/path/stream5: int16_1 <a new stream>
            """
            [Main]
              name = stream5
              path = /new/path
              datatype = int16
            [Element1]
              name=1
            """,
            # /test/deeper/stream2: float32_1 <conflicting layout>
            """
            [Main]
              name = stream2
              path = /test/deeper
              datatype = float32
            [Element1]
              name = 1
            """,
            # /invalid path//invalid: int64_1 <invalid config (ignored)>
            """
            [Main] 
              name = invalid
              path = /invalid path//
              datatype = int16
              keep = all
            [Element1]
              name=e1
            """,
        ]
        with tempfile.TemporaryDirectory() as conf_dir:
            # write the configs in 0.conf, 1.conf, ...
            i = 0
            for conf in configs:
                with open(os.path.join(conf_dir, "%d.conf" % i), 'w') as f:
                    f.write(conf)
                i += 1
            with self.assertLogs(logger=logger, level=logging.ERROR) as logs:
                load_streams.run(conf_dir, self.db)
                # log the bad path error
                self.assertRegex(logs.output[0], 'path')
                # log the incompatible layout error
                self.assertRegex(logs.output[1], 'layout')

        # now check the database:
        # should have 3 streams
        self.assertEqual(self.db.query(DataStream).count(), 5)
        # and 7 elements (orphans eliminated)
        self.assertEqual(self.db.query(Element).count(), 10)
        # Check stream merging
        #   stream1 should have a new keep value
        stream1: DataStream = self.db.query(DataStream).filter_by(name="stream1").one()
        self.assertEqual(stream1.keep_us, DataStream.KEEP_ALL)
        #   its elements should have new attributes
        self.assertEqual(stream1.elements[0].name, 'new_e1')
        self.assertEqual(stream1.elements[0].display_type, Element.DISPLAYTYPE.DISCRETE)
        self.assertIsNone(stream1.elements[0].default_min)
        self.assertEqual(stream1.elements[1].name, 'new_e2')
        self.assertEqual(stream1.elements[1].display_type, Element.DISPLAYTYPE.EVENT)
        self.assertIsNone(stream1.elements[1].default_min)
        self.assertEqual(stream1.elements[2].name, 'new_e3')
        self.assertEqual(stream1.elements[2].default_min, -10)
        # Check unconfigured streams are unchanged
        #   /test/deeper/stream2 should be the same
        stream2: DataStream = self.db.query(DataStream).filter_by(name="stream2").one()
        self.assertEqual(stream2.layout, 'int16_2')
        #   /test/deeper/stream3 should be the same
        stream3: DataStream = self.db.query(DataStream).filter_by(name="stream3").one()
        self.assertEqual(stream3.layout, 'int16_2')
        # Check new streams are added
        stream4: DataStream = self.db.query(DataStream).filter_by(name="stream4").one()
        self.assertEqual(stream4.layout, 'int16_2')

        # Check the folder structure
        # -root
        #  -test
        #    -[stream1]
        #    -deeper
        #      -[stream2]
        #  -new
        #    -path
        #      -[stream3]
        self.assertEqual(len(root.children), 2)
        for f in root.children:
            if f.name == 'test':
                self.assertEqual(len(f.data_streams), 1)
                self.assertEqual(f.data_streams[0].name, 'stream1')
                self.assertEqual(len(f.children), 1)
                deeper = f.children[0]
                self.assertEqual(deeper.name, "deeper")
                self.assertEqual(len(deeper.children), 0)
                self.assertEqual(len(deeper.data_streams), 2)
                self.assertEqual(deeper.data_streams[0].name, 'stream2')
                self.assertEqual(deeper.data_streams[1].name, 'stream3')
            elif f.name == 'new':
                self.assertEqual(len(f.data_streams), 0)
                self.assertEqual(len(f.children), 1)
                path = f.children[0]
                self.assertEqual(path.name, "path")
                self.assertEqual(len(path.children), 0)
                self.assertEqual(len(path.data_streams), 2)
                for stream in path.data_streams:
                    self.assertIn(stream.name, ['stream4', 'stream5'])
            else:
                self.fail("unexpected name: " + f.name)
